"""
sequencers
===========

Provide a generic sequencer class, as an interface for dedicated sequencer
sub-classes. This allows imsi to interface with multiple underlying sequencers.
"""
from abc import ABC, abstractmethod
from imsi.config_manager.config_manager import Configuration

def create_sequencer(seq_name):
    """
    Create and return a sequencer object
    """
    if seq_name == 'iss':
        from imsi.sequencer_interface.iss_cap import IMSISimpleSequencerInterface
        return IMSISimpleSequencerInterface()
    elif seq_name == 'maestro':
        from imsi.sequencer_interface.maestro_cap import MaestroSequencerInterface
        return MaestroSequencerInterface()
    else:
        raise ValueError(f"Specified sequencer {seq_name} not supported in imsi.")

class Sequencer(ABC):
    """
    A class that absracts the definition of a sequencer cap for imsi.

    Sub-classes for specific sequencers will provide the concerete
    implementations of the methods below, using the imsi Configuration
    information, and using their own specifically required methods.

    The sequencer specific implementation must expose back, through this
    interface, all the methods below.
    """
    @abstractmethod
    def setup(self, configuration: Configuration):
        """
        Any steps needed to setup the sequencer, including
        cloning source and creating any directories
        """
        raise NotImplementedError("sequencer config method must be defined")

    @abstractmethod
    def config(self, configuration: Configuration):
        """
        Steps needed to configure sequencer files based on updated imsi
        input. Examples might include editing resource files.
        """
        raise NotImplementedError("sequencer config method must be defined")

    @abstractmethod
    def submit(self, configuration: Configuration):
        """
        Steps needed to configure sequencer files based on updated imsi
        input. Examples might include editing resource files.
        """
        raise NotImplementedError("sequencer submit method must be defined")

    @abstractmethod
    def status(self, configuration: Configuration):
        """
        Steps needed to configure sequencer files based on updated imsi
        input. Examples might include editing resource files.
        """
        raise NotImplementedError("sequencer status method must be defined")



from imsi.tools.ensemble.table_utils.table_utils import convert_to_bracket_notation, colour
from imsi.cli.core_cli import setup, config, build, submit, save_restarts, status
from imsi.user_interface.setup_manager import InvalidSetupConfig
from imsi.utils.general import change_dir


import click
import re
from omegaconf import OmegaConf, DictConfig
from typing import Any, Generator
from pathlib import Path
import warnings


def loop_click_command_on_table(command: callable, run_path: Path, tables) -> None:
    """Calls a click cli command for each member in the table."""
    for member in tables:
        member = OmegaConf.create(member)
        runid_path = Path(run_path, member.setup.runid)
        with change_dir(runid_path):
            ctx = click.Context(command, obj={"FORCE": True})
            try:
                ctx.invoke(command)
            except Exception as e:
                click.echo(
                    f"Error while running {command.name} for {member.setup.runid}: {e}"
                )
                raise e


def save_restarts_ensemble(ensemble_config: DictConfig, tables: list) -> None:
    """Save restarts for each member in the ensemble."""
    run_path = Path(ensemble_config.ensemble_level.run_directory).resolve()
    loop_click_command_on_table(save_restarts, run_path, tables)


def compile_ensemble(ensemble_config: DictConfig, tables: list) -> None:
    """Compile each member in the ensemble."""
    run_path = Path(ensemble_config.ensemble_level.run_directory).resolve()
    loop_click_command_on_table(build, run_path, tables)


def submit_ensemble(ensemble_config: DictConfig, tables: list) -> None:
    """Submit each member in the ensemble."""
    run_path = Path(ensemble_config.ensemble_level.run_directory).resolve()
    loop_click_command_on_table(submit, run_path, tables)


def status_ensemble(ensemble_config: DictConfig, tables: list) -> None:
    run_path = Path(ensemble_config.ensemble_level.run_directory).resolve()
    loop_click_command_on_table(status, run_path, tables)


def setup_ensemble_members(run_path: Path, tables: list) -> None:
    """Run setup command for each run in the ensemble.
    Extracts the setup arguments from the table and runs the setup command.
    """
    print(f"Setting up ensemble members in {run_path}")
    with change_dir(run_path.resolve()):  # TODO: Rethink/refactor
        for member in tables:
            member = OmegaConf.create(member)
            setup_args = get_command_args(member.setup, setup)
            ctx = click.Context(setup, obj={"FORCE": True})
            try:
                ctx.invoke(setup, **setup_args)
            except InvalidSetupConfig as e:
                raise InvalidSetupConfig(
                    f"Error while running setup for {setup_args['runid']}: {e}"
                )


def check_versions_not_unique(tables: list) -> None:
    """If share_repo is True, all runs must have the same version."""
    versions = {member.setup["ver"] for member in tables if "ver" in member.setup}
    if len(versions) > 1:
        raise ValueError("All runs must have the same version if share_repo is True")


def override_table_values_for_shared_repo(run_path: Path, tables: list) -> None:
    """Override these setup args table values for shared repo."""
    for member in tables[1:]:
        setup_config = OmegaConf.create({
                "setup": {
                    "fetch_method": "link",
                    "repo": str(run_path / tables[0]["setup"]["runid"] / "src"),
                }
            })
        # Remove ver. It will use the linked version from the first run.
        member.setup.pop("ver", None)
        # Update the member's setup with the shared repo values in place
        member.setup.merge_with(setup_config.setup)


def run_setup_and_config(ensemble_config: DictConfig, tables: list) -> None:
    """Setup the ensemble run directories and run the config command."""
    run_path = Path(ensemble_config.ensemble_level.run_directory).resolve()
    run_path.mkdir(parents=True, exist_ok=True)
    if ensemble_config.ensemble_level.share_repo:
        check_versions_not_unique(tables)
        # start with first table as template
        setup_ensemble_members(run_path, [tables[0]])
        # link, set src, and ver None for remaining runs
        override_table_values_for_shared_repo(run_path, tables)
        # proceed with the rest of the runs
        setup_ensemble_members(run_path, tables[1:])
    else:
        setup_ensemble_members(run_path, tables)

    config_ensemble_members(run_path, tables)


def update_state_at_keypath(
    run_path: Path, member: DictConfig, setup_args: DictConfig
) -> None:
    """Update member resolved user config file"""
    runid = member.setup.runid
    path_to_imsi_config = Path(
        run_path, runid, f"imsi_configuration_{runid}.yaml"
    )
    imsi_config = OmegaConf.load(path_to_imsi_config)

    # isolate keys that aren't in setup to update
    for full_key_path, value in get_keys_not_in_setup_args(member, setup_args):
        update_config_at_keypath(imsi_config, full_key_path, value)

    config_yamlable = OmegaConf.to_container(imsi_config)
    with open(path_to_imsi_config, "w") as f:
        OmegaConf.save(config_yamlable, f)


def run_config(ensemble_config: DictConfig, tables: list) -> None:
    """Run the config command for each member in the ensemble."""
    run_path = Path(ensemble_config.ensemble_level.run_directory).resolve()
    config_ensemble_members(run_path, tables)


def config_ensemble_members(run_path: Path, tables: list) -> None:
    """Loop through table members and run the config command."""
    with change_dir(run_path.resolve()):  # TODO: Rethink/refactor
        for member in tables:
            member = OmegaConf.create(member)
            setup_args = get_command_args(member.setup, setup)
            update_state_at_keypath(run_path, member, setup_args)

            with change_dir(run_path / member.setup.runid):  # TODO: Rethink/refactor
                ctx = click.Context(config, obj={"FORCE": True})
                try:
                    ctx.invoke(config)
                except AttributeError as e:
                    click.echo(
                        f"Error while running config for {setup_args['runid']}."
                        "Check that the full keypaths are compatible with the configuration.\n"
                        "A common cause of this error is a missing level or key in the configuration."
                    )
                    raise e


def get_keys_not_in_setup_args(
    member: DictConfig, setup_args: DictConfig, parent_key: str = ""
) -> Generator[tuple[str, Any], None, None]:
    """
    Recursively yield key paths and values from member dictionaries
    where keys are not in `setup_args`.
    """

    for key, value in member.items():
        if key == "setup":
            # Skip the setup key itself
            continue

        full_key = f"{parent_key}:{key}" if parent_key else key
        full_key = convert_to_bracket_notation(full_key)
        if isinstance(value, (dict, DictConfig)):
            # Recursively check nested dictionaries
            yield from get_keys_not_in_setup_args(value, setup_args, full_key)
        else:
            # Yield the key path if the key is not in setup_args
            if full_key not in setup_args:
                yield full_key, value


def get_command_args(run: DictConfig, command: click.Command) -> DictConfig[str, Any]:
    """Get the command arguments from the table for the given command."""
    expected_args = {param.name: param for param in command.params}
    command_args = {k: v for k, v in run.items() if k in expected_args}
    return OmegaConf.create(command_args)


def parse_bracket_key_path(key_path: str):
    """Parse bracketed key paths like 'components[CanAM][input_files][GHG_SCENARIO]' into a list of keys."""
    pattern = r'([^\[\]]+)|\[([^\[\]]+)\]'
    matches = re.findall(pattern, key_path)
    return [m[0] or m[1] for m in matches]


def get_value_by_key_path(obj: dict, key_path: str):
    keys = parse_bracket_key_path(key_path)
    for key in keys:
        if isinstance(obj, dict):
            if key in obj:
                obj = obj[key]
            else:
                return None
        else:
            return None
    return obj


def key_exists(config: DictConfig, key_path: str) -> bool:
    keys = parse_bracket_key_path(key_path)
    parent_keys = keys[:-1]
    last_key = keys[-1]

    raw_dict = OmegaConf.to_container(config, resolve=False)
    parent = get_value_by_key_path(raw_dict, "".join(f"[{k}]" for k in parent_keys)) if parent_keys else raw_dict

    return isinstance(parent, dict) and last_key in parent


def update_config_at_keypath(config: DictConfig, key_path: str, value: Any) -> None:
    """Update the configuration dict at a nested key path."""

    if not key_exists(config, key_path):
        warnings.warn(
            colour(
                f"\n  Key path '{key_path}' is specified in the configuration but is not an existing key in the resolved imsi configuration. Please ignore this warning if this is intended.",
                "yellow"
            ),
            UserWarning
        )

    OmegaConf.update(config, key_path, value)



import pytest
from pathlib import Path
from pydantic import ValidationError
from imsi.config_manager.schema.model import Model
from imsi.config_manager.schema.components import Component, ComponentResources


# Helper fixture for valid ComponentResources
@pytest.fixture
def valid_component_resources():
    return {'mpiprocs': 4, 'ompthreads': 8}


# Helper fixture for a valid Component
@pytest.fixture
def valid_component(valid_component_resources):
    return {
        'exec': 'test_exec',
        'resources': valid_component_resources,
        'config_dir': 'valid_config',
        'namelists': {'namelist1': {}},
        'compilation': {'comp1': {}},
        'input_files': {'input1': 'input_path'},
        'output_files': {'output1': 'output_path'},
    }


# Mocking filesystem presence for config_dir validator
@pytest.fixture(autouse=True)
def mock_path_exists(monkeypatch):
    def mock_exists(path):
        # Simulate that all paths exist except "bad_dir"
        return path != Path('src', 'imsi-config', 'bad_dir')

    monkeypatch.setattr(Path, 'exists', mock_exists)


def test_component_resources_validation(valid_component_resources):
    """Test ComponentResources validation"""
    comp_res = ComponentResources(**valid_component_resources)
    assert comp_res.mpiprocs == 4
    assert comp_res.ompthreads == 8


def test_component_validation(valid_component):
    """Test Component validation with valid input"""
    comp = Component(**valid_component)
    assert comp.exec == 'test_exec'
    assert comp.config_dir == 'valid_config'


def test_component_missing_exec(valid_component):
    """Test Component validation when 'exec' is missing"""
    invalid_component = valid_component.copy()
    del invalid_component['exec']

    with pytest.raises(ValidationError):
        Component(**invalid_component)


def test_component_invalid_config_dir():
    """Test Component fails if config_dir does not exist"""
    with pytest.raises(FileNotFoundError):
        Component(
            exec='test_exec',
            resources={'mpiprocs': 4, 'ompthreads': 8},
            config_dir='bad_dir',
        )


def test_model_validation(valid_component):
    """Test Model validation with at least one component"""
    model = Model(
        name='test_model',
        source_id='test_source',
        variant_id='v1',
        short_name='tm',
        description='Test model',
        postproc_profile='default',
        components={'atm': valid_component},
    )
    assert model.name == 'test_model'
    assert 'atm' in model.components



import os
import copy
from omegaconf import OmegaConf
from pathlib import Path
import subprocess
import shlex
import shutil
from typing import Dict
from pydantic import BaseModel, field_validator
from typing import Iterable

from imsi.config_manager import config_manager as cm
from imsi.sequencer_interface.sequencers import create_sequencer
from imsi.shell_interface import shell_interface_manager
from imsi.shell_interface.config_hooks_manager import call_hooks
from imsi.utils.dict_tools import parse_vars, update, load_json
from imsi.user_interface.ui_utils import save_setup_configuration, load_run_config, apply_options_overrides
from imsi.utils.general import is_broken_symlink, _return_with_message
from imsi.utils.git_tools import is_repo_clean
from imsi import __version__


def validate_version_reqs(source_config_path: Path = Path("src/imsi-config"), version_req_file: str = "version_requirements.yaml" ):
    """
    Checks the version requirements contained in the version controlled
    config files.

    The imsi minor version is toggled when there are config breaking changes, as such we
    only require the major and minor version match.

    TODO:
        - update output to better inform users on how to find/build the right environments
    """
    path_to_version_req_file = source_config_path/version_req_file
    current_version_no_patch = ".".join(__version__.split(".")[:2])

    # check that req file exists
    if not path_to_version_req_file.exists():
        raise FileNotFoundError(f"{path_to_version_req_file} doesn't exist! This is likely because your repo hasn't been setup"
                         f" to work with {current_version_no_patch}. Please update your config files or use an older version of imsi.")

    required_version = OmegaConf.load(path_to_version_req_file)["imsi_version_requirements"]

    # confirm major/minor version match
    if current_version_no_patch != required_version:
        raise ValueError(
            f"""IMSI VERSION MIS-MATCH! Your source repo's config files are setup to use
                -> {required_version}.* <-
            But you are using
                -> {__version__} <-
            The Major and Minor version must match!

            See https://imsi.readthedocs.io/en/main/config_breaking_changes.html for more information"""
        )


def create_imsi_configuration(
    imsi_config_path: str, setup_params: Dict
) -> (cm.Configuration, cm.ConfigDatabase):
    """Build and return configuration instance and config db given imsi_config_path"""
    if not os.path.isdir(imsi_config_path):
        raise FileNotFoundError(
            f"Could not find imsi config directory at: {imsi_config_path}"
        )

    # create a DB / cm
    db = cm.database_factory(imsi_config_path)
    config_manager = cm.ConfigManager(db)

    configuration = config_manager.create_configuration(**setup_params)
    # Save the configuration for future editing/reference
    save_setup_configuration(configuration, save_config=True)

    # Create a configuration based on user input from setup cli
    return (configuration, db)


def build_run_config_on_disk(
    configuration: cm.Configuration, db: cm.ConfigDatabase, track=True, force=False
):
    """This actually creates the physical config directory on disk, and extracts/modifies various relevant files"""
    # Build the actual config directory with contents for this configuration

    shell_interface_manager.build_config_dir(db, configuration, track=track, force=force)

    # Do scheduler/sequencer setup
    sequencer = create_sequencer(configuration.setup_params.sequencer_name)
    sequencer.setup(configuration, force=force)
    sequencer.config(configuration, force=force)

    # Do other config hooks (only if constraints are met as defined in current
    # config, which will be checked via call_hooks)
    call_hooks(configuration, "post-config", force=force)

    if track:
        run_config_path = configuration.get_unique_key_value('run_config_path')
        # hooks may have changed contents of config folder - track these
        clean, _ = is_repo_clean(run_config_path)
        if not clean:
            subprocess.run(shlex.split('git add -A'), cwd=run_config_path)
            subprocess.run(shlex.split('git commit -q -m "IMSI: config_hooks:post-hook"'), cwd=run_config_path)


def reload_config_from_source(force=False):
    """
    Build a new config directory from upstream imsi source

    This will re-extract everthing out of the cloned repository to re-create
    the config directory. I.e. if one made changes in the repo after setup, and
    wanted to apply them, they would call this update function.
    """
    original_configuration = load_run_config()

    # This will rebuild the config directory completely
    imsi_config_path = original_configuration.get_unique_key_value("imsi_config_path")
    new_configuration, db = create_imsi_configuration(
        imsi_config_path,
        original_configuration.setup_params.model_dump(),
    )
    build_run_config_on_disk(new_configuration, db, force=force)


def update_config_from_state(force=False):
    """Apply changes made in the "imsi_configuration_${runid}" state file
    to the configuration and update the config directory as appropriate.
    """
    # 1. Save configuration from the configuration object to .imsi...
    # 2. Load the configuration from the .imsi... file

    user_facing_configuration = load_run_config(serialized=False)

    save_setup_configuration(user_facing_configuration, save_config=False) # no need to instantly save again
    state_configuration = load_run_config()
    # This will rebuild the config directory completely based on what is in the configuration file.
    # It would be good if there were nominal validity testing.
    db = cm.database_factory(
        state_configuration.get_unique_key_value("imsi_config_path")
    )
    build_run_config_on_disk(state_configuration, db, force=force)


class Override(BaseModel):
    options: str

    @field_validator("options")
    def validate_options(cls, v):
        if "/" not in v:
            raise ValueError("Option must be in the format <group>/<option>")
        return v


def parse_override_options(options: Iterable[str]) -> list[dict]:
    return [Override(options=o).model_dump() for o in options]


def parse_override(options: Iterable[str], force=False):
    """Parse and apply command-line option overrides to the current IMSI run configuration.

    Parameters:
    - options: An iterable of option identifier strings. Each string must contain
        a '/' separating the group and option name (e.g. "group/option" or
        "group/option=value").
    - force: If True, forces rebuilding of on-disk run artifacts
        and other actions performed by build_run_config_on_disk even if not strictly
        necessary. Defaults to False.
    """

    # get existing simulation config
    configuration = load_run_config()
    db = cm.database_factory(configuration.get_unique_key_value("imsi_config_path"))

    print(f"Overriding configuration with: {options}")
    options_list = parse_override_options(options)

    configuration_with_overrides = apply_options_overrides(
        config_dictionary=configuration.model_dump(), options=options_list
    )
    new_configuration = cm.Configuration(**configuration_with_overrides)

    # Update the saved configuration file accordingly
    # (including triggering rebuilding of /sequencer folder too, running
    # hooks, etc):
    build_run_config_on_disk(new_configuration, db, force=force)

    # Update the saved configuration file accordingly
    save_setup_configuration(new_configuration, save_config=True)


def set_selections(parm_file=None, selections=None, force=False):
    """Parse key=value pairs of selection given on the command line
    Try to apply these to the imsi selections for the sim.
    """

    # get existing simulation config
    configuration = load_run_config()

    db = cm.database_factory(configuration.get_unique_key_value("imsi_config_path"))
    config_manager = cm.ConfigManager(db)

    updated_setup_params = copy.deepcopy(configuration.setup_params)

    if parm_file:
        file_values = load_json(
            parm_file
        )  # would actually be more valuable for options I think, since
        # selections are few bu options possibly many.
        # This is updating the ._config dict in place
        updated_setup_params = update(updated_setup_params, file_values)
    if selections:
        print(f"set selections: {selections}")
        values = parse_vars(selections, none_as_str=False)
        setup_params = configuration.setup_params

        # selections must match imsi setup cli options to match parts of config
        # (some but not all allowed)
        updated_setup_params.model_name = values.pop('model', setup_params.model_name)
        updated_setup_params.experiment_name = values.pop('exp', setup_params.experiment_name)
        updated_setup_params.machine_name = values.pop('machine', setup_params.machine_name)
        updated_setup_params.compiler_name = values.pop('compiler', setup_params.compiler_name)
        updated_setup_params.sequencer_name = values.pop('sequencer', setup_params.sequencer_name)
        updated_setup_params.flow_name = values.pop('flow', setup_params.flow_name)
        updated_setup_params.postproc_profile = values.pop('postproc', setup_params.postproc_profile)

        # warn for bad selections, and don't add them to setup params
        # TODO: would be better to do this via cli (early)
        for k,v in values.items():
            print(f"**WARNING**: selection '{k}={v}' not in setup params; not added to configuration")

    # Create a new simulation that we imbue with these properties
    new_configuration = config_manager.create_configuration(**updated_setup_params.model_dump())

    # Update the saved configuration file accordingly
    # (including triggering rebuilding of /sequencer folder too, running
    # hooks, etc):
    build_run_config_on_disk(new_configuration, db, force=force)

    # Update the saved configuration file accordingly
    save_setup_configuration(new_configuration, save_config=True)

def compile_model_execs(args, force=False):
    """
    Builds all component executables by calling an upstream script from the
    repository. Should be under /bin but not enforceable.
    """
    configuration = load_run_config()
    work_dir = configuration.get_unique_key_value("work_dir")
    runid = configuration.setup_params.runid

    if not os.path.isdir(work_dir):
        raise FileNotFoundError(
            f"Could not find the run working directory at: {work_dir}"
        )

    bin_dir = os.path.join(work_dir, 'bin')
    real_bin_dir = os.path.realpath(bin_dir)

    exe_storage_dir = configuration.machine.exe_storage_dir   # path|None

    if exe_storage_dir is None:
        real_exe_storage_dir = None
    else:
        exep_exe = os.path.expandvars(os.path.realpath(exe_storage_dir))
        if exep_exe == real_bin_dir:
            # pointing to itself
            exe_storage_dir = None
            real_exe_storage_dir = None
        else:
            # exe_storage_dir/{runid}/bin <- bin
            real_exe_storage_dir = os.path.join(
                os.path.expandvars(os.path.realpath(exe_storage_dir)),
                runid,
                'bin'
                )

    # nothing to build:
    NTB_MESSAGE = "Nothing to build. To clear the executable folder(s), try: imsi -f build"

    if force:
        # cleanup all bin folders
        if os.path.exists(real_bin_dir):
            if os.path.islink(real_bin_dir):
                os.path.unlink(real_bin_dir)
            else:
                shutil.rmtree(real_bin_dir)
        if (exe_storage_dir is not None):
            if os.path.exists(real_exe_storage_dir):
                shutil.rmtree(real_exe_storage_dir)

    # setup the bin folders - local (in work_dir) and storage (exe_storage_dir as the base)
    # and account for:
    #   - changing between inputs of exe_storage_dir and
    #   - re-running imsi build (preventing full builds when not needed)
    if (exe_storage_dir is None) or (real_bin_dir == real_exe_storage_dir):
        if is_broken_symlink(bin_dir):
            # self-pointing case
            os.unlink(bin_dir)
            if not os.path.exists(real_exe_storage_dir):
                os.makedirs(real_exe_storage_dir)
                os.symlink(real_exe_storage_dir, bin_dir)
            # build required
        elif os.path.exists(bin_dir):
            if os.path.islink(bin_dir) and (real_exe_storage_dir != real_bin_dir):
                # switch from ln to dir
                os.unlink(bin_dir)
                os.makedirs(bin_dir)
                # build required
        else:
            os.makedirs(bin_dir)
            # build required
    else:
        # linking required
        if is_broken_symlink(bin_dir):
            os.unlink(bin_dir)
        else:
            if os.path.exists(real_exe_storage_dir):
                if os.path.exists(bin_dir):
                    if os.path.islink(bin_dir):
                        return _return_with_message(NTB_MESSAGE)
                    else:
                        # relink (eg previously wasn't linked to storage)
                        shutil.rmtree(bin_dir)
                        os.symlink(real_exe_storage_dir, real_bin_dir)
                        return _return_with_message(NTB_MESSAGE)
                else:
                    if os.path.exists(real_exe_storage_dir):
                        os.symlink(real_exe_storage_dir, real_bin_dir)
                        return _return_with_message(NTB_MESSAGE)
                    else:
                        # switching or remake ln
                        os.makedirs(real_exe_storage_dir)
                        os.symlink(real_exe_storage_dir, real_bin_dir)
                        # build required
            elif real_exe_storage_dir == real_bin_dir:
                return _return_with_message(NTB_MESSAGE)
            else:
                os.makedirs(real_exe_storage_dir, exist_ok=True)
                if os.path.islink(bin_dir):
                    os.unlink(bin_dir)
                if os.path.exists(bin_dir) and os.path.isdir(bin_dir):
                    shutil.rmtree(bin_dir)
                os.symlink(real_exe_storage_dir, bin_dir)

    comp_script_basename = 'imsi-tmp-compile.sh'
    comp_script = os.path.join(work_dir, comp_script_basename)

    if not os.path.exists(comp_script):
        raise FileNotFoundError(
            f"Could not find compilation file {comp_script_basename} at: {work_dir}"
        )
    compile_task = subprocess.Popen([os.path.join('.', comp_script_basename)] + list(args), cwd=work_dir)
    compile_task.wait()
    streamdata = compile_task.communicate()[0]
    rc = compile_task.returncode
    if rc != 0:
        raise ValueError(f"Error: Compiling failed with {comp_script}")


def submit_run():
    """Instantiate the configuration object and submit job to queue"""
    configuration = load_run_config()
    setup_params = configuration.setup_params
    sequencer = create_sequencer(setup_params.sequencer_name)
    sequencer.submit(configuration)


def save_restarts(args):
    """Execute the save restarts script"""

    if Path("./save_restart_files.sh").exists():
        p = subprocess.Popen(["./save_restart_files.sh"] + list(args))
        p.wait()
    else:
        raise FileNotFoundError(
            "Could not find the save_restart_files.sh script. Are you in the correct directory?"
        )

def tapeload_rs(args):
    """Execute the tapeload rs script"""

    if Path("./tapeload_rs.sh").exists():
        p = subprocess.Popen(["./tapeload_rs.sh"] + list(args))
        p.wait()
    else:
        raise FileNotFoundError(
            "Could not find the tapeload_rs.sh script. Are you in the correct directory?"
        )

def get_sequencer_status():
    configuration = load_run_config()
    setup_params = configuration.setup_params
    sequencer = create_sequencer(setup_params.sequencer_name)
    sequencer.status(configuration, setup_params)


# WIP
def query_time():
    """Instantiate the configuration / SimulationTime instances and enable querying timers"""
    configuration = load_run_config()
    sequencing_config = configuration.sequencing



import yaml
from unittest.mock import patch

import pytest

from imsi.user_interface.ui_utils import apply_options_overrides   # adjust your import


def create_option_file(tmp_path, parent_dir, file_name, content):
    """Create an option YAML file under the correct directory structure."""
    folder = tmp_path / "src" / "imsi-config" / parent_dir
    folder.mkdir(parents=True, exist_ok=True)
    filepath = folder / f"{file_name}.yaml"
    with open(filepath, "w") as f:
        yaml.safe_dump(content, f)
    return filepath


@pytest.fixture
def base_config(tmp_path):
    """A base config dict with a valid setup_params.work_dir path."""
    return {
        "model": {"name": "original"},
        "setup_params": {"work_dir": tmp_path},
    }


def test_apply_single_option(base_config):
    """Applying one override should update the dict."""
    override_content = {"model": {"name": "updated"}}
    create_option_file(
        base_config["setup_params"]["work_dir"],
        parent_dir="p1",
        file_name="opt1",
        content=override_content,
    )

    defaults = [{"p1": "opt1"}]

    # simple update: merge dicts shallowly
    def fake_update(orig, new):
        orig.update(new)
        return orig

    with patch("imsi.user_interface.ui_utils.update", side_effect=fake_update) as mock_update:
        result = apply_options_overrides(base_config, defaults)

    assert result["model"]["name"] == "updated"
    mock_update.assert_called_once()


def test_apply_multiple_options(base_config):
    override1 = {"model": {"name": "first"}}
    override2 = {"foo": "bar"}

    create_option_file(base_config["setup_params"]["work_dir"], "p1", "o1", override1)
    create_option_file(base_config["setup_params"]["work_dir"], "p2", "o2", override2)

    defaults = [{"p1": "o1"}, {"p2": "o2"}]

    # merge dictionaries
    def fake_update(orig, new):
        orig.update(new)
        return orig

    with patch("imsi.user_interface.ui_utils.update", side_effect=fake_update) as mock_update:
        result = apply_options_overrides(base_config, defaults)

    assert result["model"]["name"] == "first"
    assert result["foo"] == "bar"
    assert mock_update.call_count == 2


def test_no_defaults_returns_same_dict(base_config):
    result = apply_options_overrides(base_config, [])
    assert result is base_config  # same dict returned (function mutates)
    assert result["model"]["name"] == "original"


def test_missing_option_file_raises(base_config):
    defaults = [{"nope": "missing"}]

    with pytest.raises(FileNotFoundError):
        apply_options_overrides(base_config, defaults)



def test_apply_does_not_support_nested_dict_items(base_config):
    """Document that nested keys are not supported."""
    override = {"a": 1}
    create_option_file(base_config["setup_params"]["work_dir"], "p", "o", override)

    defaults = [{"p": "o", "nested": "ignored"}]
    with patch("imsi.user_interface.ui_utils.update", return_value=base_config.copy()) as mock_update:
        # Only first key (p) should be used
        apply_options_overrides(base_config, defaults)
        mock_update.assert_called_once()



import pytest
from pathlib import Path
from unittest.mock import patch
from imsi.tools.list.list_cli import load_rc, get_repo_paths


@pytest.fixture
def mock_env(monkeypatch, tmp_path):
    """Fixture to set up fake site/user rc files and environment variables."""
    site_file = tmp_path / "imsi.site.rc"
    user_file = tmp_path / "imsi.user.rc"

    site_file.write_text("VAR1=site_value\nVAR2=multi:val:ue\n")
    user_file.write_text("VAR1=user_value\n")

    monkeypatch.setenv("HOME", str(tmp_path))
    monkeypatch.setenv("VAR1", "env_value")
    monkeypatch.setenv("VAR3", "only_env")

    return site_file, user_file


def test_load_rc_merges_and_splits(mock_env):
    site_file, user_file = mock_env

    with patch("imsi.tools.list.list_cli.files") as mock_files, \
         patch("imsi.tools.list.list_cli.dotenv_values") as mock_dotenv:
        # fake return values for dotenv_values
        def dotenv_values_side_effect(dotenv_path):
            if str(dotenv_path).endswith("imsi.site.rc"):
                return {"VAR1": "site_value", "VAR2": "multi:val:ue"}
            if str(dotenv_path).endswith("imsi.user.rc"):
                return {"VAR1": "user_value"}
            return {}

        mock_dotenv.side_effect = dotenv_values_side_effect
        mock_files.return_value.joinpath.return_value = site_file

        result = load_rc("VAR2")
        assert ("multi", "imsi.site.rc") in result
        assert ("val", "imsi.site.rc") in result
        assert ("ue", "imsi.site.rc") in result
        # ensures no colons left
        assert all(":" not in v for v, _ in result)


def test_load_rc_returns_none_if_not_found(mock_env):
    with patch("imsi.tools.list.list_cli.files") as mock_files, \
         patch("imsi.tools.list.list_cli.dotenv_values", return_value={}):
        mock_files.return_value.joinpath.return_value = mock_env[0]
        result = load_rc("MISSING_VAR")
        assert result is None


def test_get_repo_paths_with_cli_arg(tmp_path):
    repo = tmp_path / "repo"
    config = repo / "imsi-config"
    config.mkdir(parents=True)

    with patch("imsi.user_interface.ui_manager.validate_version_reqs") as mock_validate:
        paths, sources, rel_path = get_repo_paths(str(repo))

    assert paths == [repo]
    assert sources == ["CLI args"]
    assert rel_path == Path("imsi-config")


def test_get_repo_paths_with_invalid_dir(tmp_path):
    bad_path = tmp_path / "not_a_repo"
    bad_path.write_text("file")

    with pytest.raises(ValueError):
        get_repo_paths(str(bad_path))


def test_get_repo_paths_with_rc(monkeypatch, tmp_path):
    repo = tmp_path / "repo"
    config = repo / "imsi-config"
    config.mkdir(parents=True)

    with patch("imsi.tools.list.list_cli.load_rc", return_value=[(str(repo), "rc_source")]), \
         patch("imsi.user_interface.ui_manager.validate_version_reqs") as mock_validate:
        paths, sources, rel_path = get_repo_paths(None)

    assert paths == [repo]
    assert sources == ["rc_source"]
    assert rel_path == Path("imsi-config")



"""
A utility that launches calls to 'hook's (functions) in
`config_hooks_collection.py` using conditions set in the config json.
"""

import os
from pathlib import Path

from imsi.config_manager.config_manager import Configuration
from imsi.shell_interface import config_hooks_collection
from imsi.utils.dict_tools import flatten, load_config_file


def get_hook_config(step):
    # for now make this a hard-coded file
    path = os.path.dirname(__file__)
    filename = os.path.join(path, 'config_hooks_collection_config.yaml')

    c = load_config_file(Path(filename))
    try:
        return c['config_hooks'][step]
    except KeyError:
        return {}


def call_hooks(configuration: Configuration, step='config', **kwargs):
    """Call all the hooks under the `step` based on current state of `Configuration`.

    The values in `Configuration` will be matched against the constraints
    (conditional requirements) defined in the config json. If all constraints
    are met, all hooks defined under the `step` are called (where hook names
    match the function names in config_hooks_collection.py)
    """

    key_sep = ":"
    config_hooks = get_hook_config(step)

    if not config_hooks:
        # no hooks defined for this step
        return

    # make sure that all the hooks (function names) exists
    for hook in config_hooks:
        try:
            fx = getattr(config_hooks_collection, hook['run'])
        except AttributeError as e:
            raise KeyError('imsi error: function hook {} does not exist'.format(hook['run'])) from e

    # validate all constraints
    # the first key must correspond to an imsi subconfig
    constraint_subconfigs_requested = [l for s in [c['constraints'].keys() for c in config_hooks] for l in s]
    for c in constraint_subconfigs_requested:
        try:
            configuration.model_dump()[c]
        except (ValueError, KeyError, AttributeError) as e:
            raise e

    # all subconfig are valid, then check:
    #   1. if config exists/is valid (if not, fail), then
    #   2. if constraints are met (if so, run hook)
    for hook in config_hooks:

        constraints = hook['constraints']
        constraints_met = []

        for subconfig_name, subconfig in constraints.items():

            # leverage flattened keys to do easy nested-dict comparisons
            flat_config = flatten(
                configuration.model_dump()[subconfig_name], sep=key_sep
            )
            flat_constraints = flatten(subconfig, sep=key_sep)

            for keypath, rule in flat_constraints.items():

                # test if exists
                try:
                    setting = flat_config[keypath]
                except KeyError:
                    raise ValueError("the following constraints are not configured in the current configuration for the config hook {}:\n{}".format(hook['run'], subconfig))

                # test if constraints met
                constraints_met.append(setting == rule)

        if all(constraints_met):
            # if all constraints are met, then run the hook.
            # *the function is called dynamically*
            hook_fx = getattr(config_hooks_collection, hook['run'])
            hook_fx(configuration, **kwargs)



import pytest
from pathlib import Path
from pydantic import ValidationError
from imsi.config_manager.schema.model import Model
from imsi.config_manager.schema.experiment import Experiment
from imsi.config_manager.schema.components import (
    Component,
    Components,
    ComponentResources,
)


# Mocking filesystem presence for config_dir validator
@pytest.fixture(autouse=True)
def mock_path_exists(monkeypatch):
    def mock_exists(path):
        # Simulate that all paths exist except "bad_dir"
        return path != Path('src', 'imsi-config', 'bad_dir')

    monkeypatch.setattr(Path, 'exists', mock_exists)


# Helper fixture for valid ComponentResources
@pytest.fixture
def valid_component_resources():
    return {'mpiprocs': 4, 'ompthreads': 8}


# Helper fixture for a valid Component
@pytest.fixture
def valid_component(valid_component_resources):
    return {
        'exec': 'test_exec',
        'resources': valid_component_resources,
        'config_dir': 'valid_config',
        'namelists': {'namelist1': {}},
        'compilation': {'comp1': {}},
        'input_files': {'input1': 'input_path'},
        'output_files': {'output1': 'output_path'},
    }


# test the merging of model and experiment components
@pytest.fixture
def valid_model():
    return {
        'name': 'test_model',
        'source_id': 'test_source',
        'variant_id': 'test_variant',
        'short_name': 'test_short',
        'description': 'Test model description',
        'postproc_profile': 'default',
        'inherits_from': ['parent_model'],
        'scientifically_validated': True,
        'repository_tag': 'v1.0',
        'uxxx': None,
        'prefix': None,
        'model_filename_prefix': None,
        'model_rs_filename_prefix': None,
    }


@pytest.fixture
def valid_experiment():
    """Fixture to provide valid experiment data."""
    return {
        'name': 'Test Experiment',
        'experiment_id': 'exp_001',
        'subexperiment_id': 'sub_001',
        'activity_id': 'act_001',
        'mip_era': 'CMIP6',
        'model_type': 'AOGCM',
        'start_time': '2025',
        'end_time': '2025',
        'parent_runid': 'parent_123',
        'parent_branch_time': '2024',
        'inherits_from': None,
        'supported_models': ['ModelA', 'ModelB'],
        'components': {'atmosphere': 'AMIP', 'ocean': 'OMIP'},
    }


def test_merged_components(valid_model, valid_experiment, valid_component):
    """Test merging of model and experiment components"""
    model = Model(**valid_model)
    model.components = valid_component
    experiment = Experiment(**valid_experiment)

    # Assuming the components are merged into the model
    merged_components = {**model.components, **experiment.components}

    assert len(merged_components) == len(model.components) + len(experiment.components)


def test_invalid_component_resources():
    """Test ComponentResources validation with invalid data"""
    with pytest.raises(ValidationError):
        ComponentResources(mpiprocs='invalid', ompthreads=8)


def test_missing_components_config_dir():
    """Test Component validation with invalid input"""
    invalid_components = {
        'CanTEST': {
            'exec': 'test_exec',
            'resources': {'mpiprocs': 4, 'ompthreads': 8},
            'namelists': {'namelist1': {}},
            'compilation': {'comp1': {}},
            'input_files': {'input1': 'input_path'},
            'output_files': {'output1': 'output_path'},
        }
    }
    with pytest.raises(FileNotFoundError):
        Components(**invalid_components)


def test_missing_config_dir():
    """Test Component validation when 'config_dir' is missing"""
    invalid_component = {
        'exec': 'test_exec',
        'config_dir': 'bad_dir',
        'resources': {'mpiprocs': 4, 'ompthreads': 8},
        'namelists': {'namelist1': {}},
        'compilation': {'comp1': {}},
        'input_files': {'input1': 'input_path'},
        'output_files': {'output1': 'output_path'},
    }

    with pytest.raises(FileNotFoundError):
        Component(**invalid_component)


def test_fail_single_component():
    """Test Component validation with minimum required fields"""
    invalid_component = {
        'exec': 'test_exec',
        'resources': {'mpiprocs': 4, 'ompthreads': 8},
        # Missing config_dir
    }

    with pytest.raises(ValidationError):
        Component(**invalid_component)


def test_fail_component_resources():
    """Test ComponentResources validation with missing fields"""
    invalid_component_resources = {
        'mpiprocs': 4,
        # Missing ompthreads
    }

    test_component = {
        'exec': 'test_exec',
        'resources': invalid_component_resources,
        'config_dir': 'valid_config',
    }

    test_components = {
        'CanTEST': test_component,
    }

    with pytest.raises(ValidationError):
        Components(**test_components)



import pytest
from pathlib import Path
from types import NoneType
import pandas as pd
from imsi.sequencer_interface.maestro_status import (
    get_root_module,
    check_entry_and_get_dataframe,
    create_maestro_dataframe,
    check_row_for_string,
    filter_df
)

# construct path relative to this file
mock_begin = Path(__file__).parent.resolve() / "mock_begin"
mock_empty = Path(__file__).parent.resolve() / "mock_empty"
mock_stop = Path(__file__).parent.resolve() / "mock_stop"
mock_end = Path(__file__).parent.resolve() / "mock_end"


@pytest.mark.parametrize("wrk_dir, expected", [
    (mock_begin, "canesm"),
    (mock_empty, "banana")
])
def test_get_root_module(wrk_dir, expected):
    assert get_root_module(wrk_dir) == expected


@pytest.mark.parametrize("root_module, entry_path, expected", [
    ("canesm", Path(mock_begin, "sequencer", "sequencing", "status", "12345678910123"), pd.DataFrame),
    ("canesm", Path(mock_end, "sequencer", "sequencing", "status", "20250224220900"), pd.DataFrame),
    ("canesm", Path(mock_stop, "sequencer", "sequencing", "status", "20250222000900"), pd.DataFrame),
    ("banana", Path(mock_empty), NoneType)
])
def test_check_entry_and_get_dataframe(root_module, entry_path, expected):
    assert isinstance(check_entry_and_get_dataframe(root_module, entry_path), expected)


@pytest.mark.parametrize("entry_path, expected", [
    (Path(mock_begin, "sequencer", "sequencing", "status", "12345678910123"), pd.DataFrame),
    (Path(mock_end, "sequencer", "sequencing", "status", "20250224220900"), pd.DataFrame),
    (Path(mock_stop, "sequencer", "sequencing", "status", "20250222000900"), pd.DataFrame),]
)
def test_create_maestro_dataframe(entry_path, expected):
    df = create_maestro_dataframe(entry_path)
    assert isinstance(df, expected)
    assert len(df) > 0
    assert len(df.columns) > 0


def test_check_row_for_string():
    df = pd.DataFrame({"col1": ["canesm", "banana", "apple"], "col2": ["canesm", "banana", "apple"]})
    assert check_row_for_string(df.iloc[0], "canesm")
    assert not check_row_for_string(df.iloc[1], "horse")


def test_filter_df():
    # Create a sample DataFrame
    data = {
        "col1": [
            "job1.+1.begin",
            "job2.+1.abort.stop",
            "job3.+1.end",
            "job4.+1.catchup",
            "job5.+1.submit",
            "job6.+1.other"  # check that it ignores non recognized statuses
        ],
        "col2": [
            "job1.+1.begin",
            "job2.+1.abort.stop",
            "job3.+1.end",
            "job4.+1.catchup",
            "job5.+1.submit",
            "job6.+1.other"
        ]
    }
    df = pd.DataFrame(data)

    # Expected DataFrame after filtering
    expected_data = {
        "col1": [
            "job1.+1.begin",
            "job2.+1.abort.stop",
            "job3.+1.end",
            "job4.+1.catchup",
            "job5.+1.submit",
        ],
        "col2": [
            "job1.+1.begin",
            "job2.+1.abort.stop",
            "job3.+1.end",
            "job4.+1.catchup",
            "job5.+1.submit",
        ],
        "job_file": [
            "job1.+1.begin",
            "job2.+1.abort.stop",
            "job3.+1.end",
            "job4.+1.catchup",
            "job5.+1.submit",
        ],
        "loop_number": [1, 1, 1, 1, 1],
        "job_base": [
            "job1",
            "job2",
            "job3",
            "job4",
            "job5"
        ],
        "job_status": [
            "begin",
            "stop",
            "end",
            "catchup",
            "submit"
        ]
    }
    expected_df = pd.DataFrame(expected_data)

    # Call the filter_df function
    filtered_df = filter_df(df)
    # Assert the filtered DataFrame is as expected
    pd.testing.assert_frame_equal(
        filtered_df.reset_index(drop=True),
        expected_df.reset_index(drop=True),
        check_dtype=False
    )



import os
from abc import ABC, abstractmethod
from typing import Dict

from imsi.utils.dict_tools import combine_yaml_configs, combine_json_configs, parse_config_inheritance

class ConfigDatabase(ABC):
    '''Defines an abstract interface for a config DB
       This might be useful if in the future different DBs could be used.
       For instance there could be an sqlite or xml based db instead of json,
       which is why the ConfigDatabase abstraction is in place, allowing future
       flexibility (Open for expansion, closed for change principle)

       The interface defines two "get" methods which return a dictionary of the
       config
    '''
    @abstractmethod
    def get_config(self, config_name: str) -> Dict:
        '''Returns all entries under a configuration category, for example
        if config_name="machines", it would return a dict containing the configurations
        for all machines in the library (which machine names as keys)'''
        pass

    @abstractmethod
    def get_parsed_config(self, config_name: str, config_key: str) -> Dict:
        '''Returns a specific named entry from a specific configuration category.
        For example if config_name="machines", and config_key="ppp6", it would \
        return a dict containing the machine configuration for ppp6'''
        pass

class YAMLConfigDatabase(ConfigDatabase):
    """A specific implementation of a ConfigDatabase, which reads the inputs from a hierachy
    of YAML files.
    """

    def __init__(self, imsi_config_path):
        self.imsi_config_path = imsi_config_path
        self.database = self._load_config()

    def _load_config(self) -> Dict:
        """Parses the imsi_config_path for all json files and combines
        them into one big dict
        """
        if not os.path.exists(self.imsi_config_path):
            raise FileNotFoundError(
                f"Could not find imsi config directory at: {self.imsi_config_path}"
            )

        return combine_yaml_configs(self.imsi_config_path)

    def get_config(self, config_name: str) -> Dict:
        """Returns all entries under a configuration category, for example
        if config_name="machines", it would return a dict containing the configurations
        for all machines in the library (which machine names as keys)"""
        if config_name not in self.database:
            raise NameError(
                f"{config_name} not found in database created from {self.imsi_config_path}"
            )
        return self.database[config_name]

    def get_parsed_config(self, config_name: str, config_key: str) -> Dict:
        """Returns a specific named entry from a specific configuration category.
        For example if config_name="machines", and config_key="ppp6", it would \
        return a dict containing the machine configuration for ppp6"""
        config_realm_data = self.get_config(config_name)
        if config_key not in config_realm_data:
            raise NameError(
                f"{config_key} not found in {config_name} in the configuration "
                f"database created from {self.imsi_config_path}"
            )
        return parse_config_inheritance(config_realm_data, config_key)

class JsonConfigDatabase(ConfigDatabase):
    '''A specific implementation of a ConfigDatabase, which reads the inputs from a hierachy
       of json files.
    '''
    def __init__(self, imsi_config_path):
        self.imsi_config_path = imsi_config_path
        self.database = self._load_config()


    def _load_config(self) -> Dict:
        '''Parses the imsi_config_path for all json files and combines
        them into one big dict
        '''
        if not os.path.exists(self.imsi_config_path):
            raise FileNotFoundError(f'Could not find imsi config directory at: {self.imsi_config_path}')

        return combine_json_configs(self.imsi_config_path)

    def get_config(self, config_name: str) -> Dict:
        '''Returns all entries under a configuration category, for example
        if config_name="machines", it would return a dict containing the configurations
        for all machines in the library (which machine names as keys)'''
        if config_name not in self.database:
            raise NameError(f'{config_name} not found in database created from {self.imsi_config_path}')
        return self.database[config_name]

    def get_parsed_config(self, config_name: str, config_key: str) -> Dict:
        '''Returns a specific named entry from a specific configuration category.
        For example if config_name="machines", and config_key="ppp6", it would \
        return a dict containing the machine configuration for ppp6'''
        config_realm_data = self.get_config(config_name)
        if config_key not in config_realm_data:
            raise NameError(f'{config_key} not found in {config_name} in the configuration '
                            f'database created from {self.imsi_config_path}')
        return parse_config_inheritance(config_realm_data, config_key)



"""
nml_tools
=========

This is a module to collect common namelist functions.

This currently exists because neither f90nml nor the rpn nml tools correctly parse
the nemo namelists. Ideally all these functions would be replaced with more
robust, community packages.

NCS, 10/2021
"""
import os
import re


def nml_update(nml_input_file, nml_output_file, nml_changes):
   """
   An interface function to update namelist parameters.

   Inputs:
      nml_input_file : str
         path to the namelist file to load in
      nml_output_file : str
         path to the updated namelist file to write
      nml_changes : dict
         A nested dict containing. At the top level the keys are the names of the namelists in nml_input_file,
         the values are dicts containing key = value pairs to change in the namelist.

   This is a basic python implementation of the `mod_nl` routine. Ideally the work would be done by a standard
   fortran namelist parser, such as f90nml. However, no existing parsers work correctly on NEMO namelists.
   """
   with open(nml_input_file, 'r') as infile:
      filedata = infile.read()

   # todo - it could be possible to preserve comments after the replacement by using a more
   # sophisticated re, or by doing a split on `!` like in the read function below.
   with open(nml_output_file, 'w') as outfile:
      for line in filedata.split('\n'):
         lineout = line
         for nam, namdict in nml_changes.items():
            for k, v in namdict.items():
               k0 = re.sub(r'(\(|\))',r'\\\1',k)
               lineout = re.sub(rf'^ *{k0} *=.*', f'{k} = {v}', lineout)
         outfile.write(f'{lineout}\n')


def nml_write(nml_output_file, nml):
   """
   An interface function to write namelists from a dict.

   Inputs:
      nml_output_file : str
         path to the namelist file to write
      nml : dict
         A nested dict containing. At the top level the keys are the names of the namelists,
         the values are dicts containing key = value pairs of the namelist.

   Ideally the work would be done by a standard fortran namelist parser, such as f90nml.
   However, no existing parsers work correctly on NEMO namelists.
   """
   with open(nml_output_file, 'w') as outfile:
      for nam, namdict in nml.items():
         outfile.write(f'&{nam}\n')
         for k, v in namdict.items():
            outfile.write(f'{k} = {v}\n')
         outfile.write(f'/\n')

def nml_read(nml_input_file):
   """
   An interface function to read namelists into a dict.

   Inputs:
      nml_input_file : str
         path to the namelist file to read

   Ideally the work would be done by a standard fortran namelist parser, such as f90nml.
   However, no existing parsers work correctly on NEMO namelists.
   """
   with open(nml_input_file, 'r') as infile:
      filedata = infile.read()

   nml = {}
   for line in filedata.split('\n'):
      if line.strip().startswith('&'):
          key = re.sub(r'!.*', '', line).replace('&','').strip()
          nml[key] = {}
          continue
          print(key)
      if line.strip().startswith('/'):
         key = 'ERROR'
         continue
      if line.strip().startswith('!'):
         continue
      if line:
         key_value_comment = line.split('!',1)
         key_value = key_value_comment[0]
         if len(key_value_comment)>1:
            comment = key_value_comment[1]
         else:
            comment = ''
         k = re.sub(r'=.*', '', key_value).strip()
         v = re.sub(r'.*=', '', key_value).strip()
         nml[key][k] = v.strip()
         #print(f'{k} = {v} !{comment}')
   return nml

def cpp_update(cpp_input_file, cpp_output_file, cpp_changes, verbose=False):
    """
    An interface function to update cpp keys.

    Inputs:
       cpp_input_file : str
          path to the cpp file to load in
       cpp_output_file : str
          path to the updated cpp file to write
       cpp_changes : dict
          key = value pairs to change in the file, where
          key appears in the default cpp file, and value is the value to replace it with

    This is a basic python implementation of the `mod_nl` routine. Ideally the work would be done by a standard
    fortran namelist parser, such as f90nml. However, no existing parsers work correctly on NEMO namelists.
    """
    # Note not yet tested on NEMO .fcm files
    with open(cpp_input_file, 'r') as infile:
        filedata = infile.read()

    with open(cpp_output_file, 'w') as outfile:
        for line in filedata.split('\n'):
            lineout = line
            if ('replace' in cpp_changes.keys()):
                for k, v in cpp_changes['replace'].items():
                    if verbose:
                        print(f'in CPP replacing {k} {v}')

                    # likely too general for fcm files, since
                    # multiple keys appear on one line.
                    lineout = re.sub(rf'.*{k}.*', f'{v}', line)

            outfile.write(f'{lineout}\n')
        if ('add' in cpp_changes.keys()):
            for k, v in cpp_changes['add'].items():
                lineout = k
                outfile.write(f'{lineout}\n')

def update_env_file(infile: str, outfile: str = None, updates: dict = None,
                   commment_char: str = '#', key_value_only: bool = True):
   """Update simple shell/environment files.

   Only replaces values of existing keys, i.e. does not add key-value
   pairs that are not already in the file.

   If `key_value_only` is True, then only lines that are in the key=value
   pattern (and commented lines) will be written to the output file.
   To keep all lines, set `key_value_only=False`.
   """
   comment_char = "#"
   linepattern = re.compile(r'(\w+)\s*=\s*(.*)')

   if outfile is None:
      outfile = infile
   if not updates:
      raise ValueError('updates must be a dict with at least one key-value pair')
      # return

   with open(infile, 'r') as f:
      filedata = f.read()

   try:
      with open(outfile, 'w') as out:
            for line in filedata.split('\n'):
               ls = line.strip()
               lineout = line

               if not line or ls.startswith(comment_char):
                  out.write(lineout+'\n')
                  continue

               m = linepattern.search(ls)
               if m:
                  param, val_current = m.groups()
                  if param in updates:
                        val_new = str(updates[param]).strip('"')
                        if val_current:
                           lineout = line.replace(val_current, val_new)
                        else:
                           # handles if the value was left unset (empty after =)
                           lineout = f'{line}{val_new}'

                  out.write(lineout+'\n')

               else:
                  if not key_value_only:
                        # keep the other lines
                        out.write(lineout+'\n')

   except Exception as e:
      if infile != outfile and os.path.exists(outfile):
            os.remove(outfile)
      raise Exception(f'Error writing file {outfile}; no file produced') from e



from pydantic import BaseModel, ConfigDict
from typing import Optional, Dict


class Utilities(BaseModel):
    """Sequencing configuration dataclass"""

    model_config = ConfigDict(extra='allow')
    files_to_extract: Optional[Dict[str, str]] = {}



from pydantic import BaseModel, ConfigDict


class PostProcessing(BaseModel):
    """Post-processing configuration dataclass"""

    # Current this is just a big dict, ingested straight from
    # what is in the database. Almost certainly though it will
    # need to be developed to make it more useful.
    # Likely, these settings need to also appear and be modifiable within
    # experiment definitions (ala components).
    model_config = ConfigDict(extra='allow')



import os
from dotenv import load_dotenv

load_dotenv(os.path.join(os.path.dirname(__file__), '.env.test'))


def pytest_addoption(parser):
    parser.addoption(
        '--target-repo',
        action='store',
        default=os.environ.get("TEST_EXT_REPO", "git@gitlab.science.gc.ca:CanESM/CanESM5.git"),
        help='Git repository that contains configs to test with.'
    )
    parser.addoption(
        '--target-branch',
        action='store',
        default=os.environ.get("TEST_VERSION", "develop_canesm"),
        help='Branch in git repository to test with.'
    )
    parser.addoption(
        '--target-model',
        action='store',
        default=os.environ.get("TEST_MODEL", "canesm"),
        help='Model to use in unit tests'
    )
    parser.addoption(
        '--target-exp',
        action='store',
        default=os.environ.get("TEST_EXP", "cmip6-piControl"),
        help='Experiment to use in unit tests'
    )
    parser.addoption(
        '--target-machine',
        action='store',
        default=os.environ.get("TEST_MACHINE", "ppp6"),
        help='Machine to use in unit tests'
    )
    parser.addoption(
        '--target-sequencer',
        action='store',
        default=os.environ.get("TEST_SEQ", "iss"),
        help='Sequencer to use in unit tests'
    )


import pytest
from pathlib import Path
from unittest.mock import patch
from pydantic import ValidationError
from imsi.user_interface.setup_manager import ValidatedSetupOptions, InvalidSetupConfig


# Fixtures
@pytest.fixture
def mock_path_exists():
    """Fixture to mock Path.exists() method."""
    with patch.object(Path, "exists", return_value=True):
        yield


@pytest.fixture
def valid_inputs(request):
    pars = {
        "repo": request.config.getoption("--target-repo"),
        "ver": request.config.getoption("--target-branch"),
        "model": request.config.getoption("--target-model"),
        "exp": request.config.getoption("--target-exp"),
        "runid": "valid-runid",
        "fetch_method": "clone",
        "seq": "sequence",
        "machine": request.config.getoption("--target-machine"),
        "flow": "flow-abc",
        "postproc": "bananas!"
    }
    return pars


# Test cases for field validation
@pytest.mark.parametrize(
    "runid, should_raise",
    [
        ("valid-runid", False),
        ("invalid_runid!", True),
        ("runid_with_uppercase", True),
        ("short", False),
        ("run-id-hyphen-09", False),
    ],
)
def test_validate_runid(valid_inputs, runid, should_raise):
    valid_inputs["runid"] = runid
    if should_raise:
        with pytest.raises(InvalidSetupConfig, match="Your runid --->"):
            ValidatedSetupOptions(**valid_inputs)
    else:
        options = ValidatedSetupOptions(**valid_inputs)
        assert options.runid == runid


@pytest.mark.parametrize(
    "repo, should_raise",
    [
        # get the valid repo from the command line args
        ("placeholder", False),
        ("git@github.com:user/repo.git", True),  # doesn't exist!
        ("invalid_repo_url", True),
        ("", True),
    ],
)
def test_validate_git_clone(valid_inputs, repo, should_raise):
    if repo == "placeholder":
        repo = valid_inputs["repo"]
    valid_inputs["repo"] = repo
    if should_raise:
        with pytest.raises(
            InvalidSetupConfig
        ):
            ValidatedSetupOptions(**valid_inputs)
    else:
        options = ValidatedSetupOptions(**valid_inputs)
        assert options.repo == repo


@pytest.mark.parametrize(
    "fetch_method, ver, should_raise",
    [
        ("clone", "develop_canesm", False),
        ("clone", None, True),
        ("link", None, False),
        ("copy", None, False),
        ("invalid_method", None, True),
    ],
)
def test_validate_fetch_method(valid_inputs, fetch_method, ver, should_raise):
    valid_inputs["fetch_method"] = fetch_method
    valid_inputs["ver"] = ver
    if should_raise:
        with pytest.raises(
            InvalidSetupConfig
        ):
            ValidatedSetupOptions(**valid_inputs)
    else:
        options = ValidatedSetupOptions(**valid_inputs)
        assert options.fetch_method == fetch_method


@pytest.mark.parametrize(
    "fetch_method, ver, should_raise",
    [
        ("clone", "develop_canesm", False),
        ("clone", None, True),
        ("link", "develop_canesm", True),
        ("copy", "develop_canesm", True),
        ("link", None, False),
        ("copy", None, False),
    ],
)
def test_error_if_version_unused_or_used(valid_inputs, fetch_method, ver, should_raise):
    valid_inputs["fetch_method"] = fetch_method
    valid_inputs["ver"] = ver

    if should_raise:
        with pytest.raises(InvalidSetupConfig):
            ValidatedSetupOptions(**valid_inputs)
    else:
        options = ValidatedSetupOptions(**valid_inputs)
        assert options.fetch_method == fetch_method
        assert options.ver == ver


# Test missing required fields
@pytest.mark.parametrize("missing_field", ["runid", "repo", "fetch_method", "exp"])
def test_missing_required_fields(valid_inputs, missing_field):
    valid_inputs.pop(missing_field, None)
    with pytest.raises(ValidationError, match=f"{missing_field}"):
        ValidatedSetupOptions(**valid_inputs)


# warning issued when sha used because default is clone == shallow clone
@pytest.mark.filterwarnings("ignore:Can't use --fetch_method:UserWarning")
@pytest.mark.parametrize(
    "ver, should_raise",
    [
        ("develop_canesm", False),
        ("bad-branch-yolo", True),  # pretty sure nobody will name a branch this ;)
        ("v5.1.6", False),
        ("68e8a27e42ab80588a08d7326af3c8a793f6f774", False),  # commit hash
        ("68e8a27", False),  # Partial commit hash
        ("??$?$?", True)  # invalid commit hash
    ],
)
def test_validate_branch(valid_inputs, ver, should_raise):
    valid_inputs["ver"] = ver
    if should_raise:
        with pytest.raises(
            InvalidSetupConfig, match="not a valid remote branch or tag"
        ):
            ValidatedSetupOptions(**valid_inputs)
    else:
        options = ValidatedSetupOptions(**valid_inputs)
        assert options.ver == ver



'''
Generates the shell interface to the modelling system.

In general this means creating a "config" directory with various files that the
model and downstream utilities ingest, such as namelists, and shell_parameters files.
'''
import os
import shlex
import subprocess

import os
import stat

from imsi.config_manager import config_manager as cm
from imsi import shell_interface as si
from imsi.config_manager import config_manager as cm
from imsi.shell_interface.config_hooks_manager import call_hooks
from imsi.utils.general import write_shell_script, write_shell_string
from imsi.utils.dict_tools import recursive_lookup
from imsi.utils.git_tools import ensure_git_config, git_add_all, get_head_hash, repo_has_commits
from imsi.user_interface.ui_utils import freeze_run_configuration

def build_config_dir(
    db: cm.ConfigDatabase, configuration: cm.Configuration, track=True, force=False
):
    """Main composition function to manange the build the config directory with appropriately parsed content"""

    # Unpacking for ease of reuse below.
    # FUTURE reconsider function args and potential coupling
    config_dict = configuration.model_dump()
    component_dict = configuration.components.model_dump()
    sequencing_config = configuration.sequencing.model_dump()
    run_dates = sequencing_config["run_dates"]
    imsi_config_path = configuration.get_unique_key_value("imsi_config_path")
    run_config_path = configuration.get_unique_key_value("run_config_path")
    work_dir = configuration.get_unique_key_value("work_dir")

    # 1. Create config directory structure
    si.shell_interface_utilities.create_config_dirs(
        run_config_path, component_dict.keys(), force=force
    )

    if track:
        # a temporary commit, before contents are added to /config
        has_commits = repo_has_commits(path=run_config_path)
        git_add_all(path=run_config_path)
        subprocess.run(shlex.split('git commit -q -m "tmp" --allow-empty'), cwd=run_config_path)
        base_sha = get_head_hash(run_config_path)

    # 2. shell_parameters files
    # Almost certainly should be done in config_manager, not here.
    # The fact that the db is used here and not elsewhere indicates this.
    # Like the db should only be used in the config not the interfaces, while the
    # configuration itself is used in the interfaces. i.e. the Configuration is the SSOT
    shell_config = si.shell_config_parameters.set_shell_config(
        db.get_config("shell_config"), config_dict
    )

    write_shell_script(
        os.path.join(run_config_path, "shell_parameters"),
        si.shell_config_parameters.generate_shell_parameters(shell_config),
    )

    # 3. Write computational env / compilation files
    # TODO refactor using common writing functions and to more explicitly
    # specify paths etc
    machine_comp_env = {
        configuration.machine.name: configuration.machine.computational_environment
    }
    for k, v in configuration.machine.site.items():
        machine_comp_env[k] = v["computational_environment"]

    basename_prefix = "computational_environment_"
    for m, c in machine_comp_env.items():
        write_shell_script(
            os.path.join(run_config_path, f"{basename_prefix}{m}"),
            si.shell_comp_environment.generate_computational_environment(
                c,
                m,
                compiler_name=configuration.compiler.name,  # !!
            ),
        )

    # write  the computational environment 'controller'
    write_shell_script(
        os.path.join(run_config_path, "computational_environment"),
        si.shell_comp_environment.generate_computational_environment_controller(
            configuration.machine, run_config_path, basename_prefix
        ),
    )

    write_shell_script(
        os.path.join(run_config_path, "compilation_template"),
        si.shell_comp_environment.generate_compilation_template(configuration),
    )

    # 4. Process namelists and cpp/compilation files
    # - take in reference versions ('templates') and modifying them according
    # to changes listed in component_config
    for component, component_config in component_dict.items():
        si.shell_inputs_outputs.process_model_config_files(
            component,
            component_config,
            imsi_config_path,
            run_config_path,
            "compilation",
            shell_config=shell_config,
        )

        si.shell_inputs_outputs.process_model_config_files(
            component,
            component_config,
            imsi_config_path,
            run_config_path,
            "namelists",
            shell_config=shell_config,
        )

    # extract identified utlitity scripts (like compile)
    files_to_extract = configuration.utilities.files_to_extract

    if not files_to_extract:  # backwards compatibility
        files_to_extract = {
            "save_restart_files.sh": "models/save_restart_files.sh",
            "imsi-tmp-compile.sh": "imsi-tmp-compile.sh",
        }
    si.shell_inputs_outputs.extract_utility_files(
        files_to_extract, imsi_config_path, work_dir
    )

    # 5. Write shell scripts for use in pre/post ludes
    # FUTURE Could split this by component and join with above for loop
    # input files (prelude)
    write_shell_script(
        os.path.join(run_config_path, "imsi_get_input_files.sh"),
        si.shell_inputs_outputs.generate_input_script_content(
            component_dict, shell_config
        ),
    )
    # directory packing in postlude
    write_shell_script(
        os.path.join(run_config_path, "imsi_directory_packing.sh"),
        si.shell_inputs_outputs.generate_directory_packing_content(component_dict),
    )
    # final saving in postlude
    write_shell_script(
        os.path.join(run_config_path, "imsi_save_output_files.sh"),
        si.shell_inputs_outputs.generate_final_file_saving_content(component_dict),
    )

    # 6. Write diag_parameters
    diag_config = config_dict["postproc"]
    write_shell_script(
        os.path.join(run_config_path, "diag_parameters"),
        si.shell_diag_parameters.generate_diag_parameters_content(diag_config),
    )

    # 7. Generate a file with the model run command
    output_path = os.path.join(run_config_path, "imsi-model-run.sh")
    run_commands = list(recursive_lookup('model_run_commands', sequencing_config))[0]
    write_shell_string(output_path, run_commands, make_executable=True)

    # 8. Write timing files
    si.shell_timing_vars.validate_timers_config(run_dates)
    write_shell_script(
        os.path.join(run_config_path, 'model_submission_job_start-stop_dates'),
        si.shell_timing_vars.generate_timers_config(run_dates,'model_submission_job')
        )

    write_shell_script(
        os.path.join(run_config_path, 'model_inner_loop_start-stop_dates'),
        si.shell_timing_vars.generate_timers_config(run_dates,'model_inner_loop')
        )

    write_shell_script(
        os.path.join(run_config_path, 'postproc_submission_job_start-stop_dates'),
        si.shell_timing_vars.generate_timers_config(run_dates,'postproc_submission_job')
        )

    # 9. call config_hooks:config
    # (constraints checked from configuration)
    call_hooks(configuration, "config", force=force)

    # 10. cp resolved configuration into /config
    freeze_run_configuration(configuration)

    # finally:
    if track:
        # create the final commit by resetting to the original hash
        # and amending *that* commit. this ensures that there is only
        # a single commit by the end of this function.
        # note: pass the path explicitly incase any of the hooks cwd
        subprocess.run(shlex.split(f'git reset -q --soft {base_sha}'), cwd=run_config_path)
        git_add_all(path=run_config_path)
        ensure_git_config()
        subprocess.run(shlex.split('git commit -q --amend -m "IMSI: config" --allow-empty'), cwd=run_config_path)

        # check if the most recent commit was empty. if so, we want
        # to revert to prev commit (and only if there is more than one
        # commit)
        proc = subprocess.run(shlex.split('git diff-tree --quiet HEAD'), cwd=run_config_path, capture_output=True)
        if proc.returncode == 0 and has_commits:
            subprocess.run(shlex.split('git reset -q --hard HEAD~1'), cwd=run_config_path)



import click
from imsi.cli.lazy import LazyGroup


class SectionedGroup(LazyGroup):
    """Like LazyGroup, but prints commands and sub‑groups in separate sections in the help message."""

    def format_commands(self, ctx, formatter):
        commands, groups = [], []

        for name in self.list_commands(ctx):
            cmd = self.get_command(ctx, name)
            if cmd is None or cmd.hidden:
                continue
            row = (name, cmd.get_short_help_str())
            # check if in a group or a command
            (groups if isinstance(cmd, click.Group) else commands).append(row)

        if commands:
            with formatter.section("Commands"):
                formatter.write_dl(commands)

        if groups:
            # Blank line between sections is automatic
            with formatter.section("Command Groups"):
                formatter.write_dl(groups)



import click
from importlib.resources import files


def add_site_repos_to_rc(path_to_site_repos):
    path_to_rc = files("imsi").joinpath("imsi.site.rc")
    with open(path_to_rc, "a") as f:
        # append IMSI_DEFAULT_CONFIG_REPOS=path_to_site_repos
        f.write(f"\nIMSI_DEFAULT_CONFIG_REPOS={path_to_site_repos}\n")
    click.echo(f"Added IMSI_DEFAULT_CONFIG_REPOS={path_to_site_repos} to {path_to_rc}")


@click.command()
@click.option('--path-to-site-repos', help='Location on disk where site-repos are stored. Will add value to the imsi.site.rc file. Not meant for regular imsi users, only for system install.', required=True)
def post_install(path_to_site_repos):
    add_site_repos_to_rc(path_to_site_repos)



import pytest
from omegaconf import OmegaConf, DictConfig
from pydantic import ValidationError
from imsi.tools.ensemble.table_utils.table_model import TextTableVars
from imsi.tools.ensemble.table_utils.data_model import (
    MemberLevelVars,
    EnsembleLevelVars,
)

@pytest.fixture
def config_params():
    """Static defaults so the tests don’t need CLI options."""
    return {
        "repo": "origin",
        "branch": "main",
        "model": "canam",
        "experiment": "piControl",
    }


@pytest.fixture
def setup_config(config_params):
    return OmegaConf.create(
        {
            "ver": config_params["branch"],
            "repo": config_params["repo"],
            "fetch_method": "clone",
            "machine": "ppp6",
            "model": config_params["model"],
        }
    )


@pytest.fixture
def config_data_one(setup_config):
    """A minimal but valid member-level configuration."""
    return OmegaConf.create(
        {"member_level": {"setup": {**setup_config, "runid": ["run1"]}}}
    )


@pytest.fixture
def text_table_basic(tmp_path):
    """setup:runid  setup:exp"""
    p = tmp_path / "basic.txt"
    p.write_text(
        "setup:runid  setup:exp\n"
        "run-01       cmip6-piControl\n"
        "run-02       cmip6-piControl\n"
    )
    return str(p)


@pytest.fixture
def text_table_nested(tmp_path):
    """one:two:three in addition to setup keys."""
    p = tmp_path / "nested.txt"
    p.write_text(
        "setup:runid  setup:exp  one:two:three\n"
        "run-01       cmip6-piControl  1\n"
        "run-02       cmip6-piControl  2\n"
    )
    return str(p)


@pytest.fixture
def text_table_conflict(tmp_path):
    """`one:two` AND `one:two:three` => conflict."""
    p = tmp_path / "conflict.txt"
    p.write_text(
        "setup:runid  one:two  one:two:three\n"
        "run-01       10       100\n"
        "run-02       20       200\n"
    )
    return str(p)


def _ensemble_cfg(path: str) -> DictConfig:
    """Create an ensemble-level DictConfig pointing at *path*."""
    return OmegaConf.create(
        {
            "user": "test_user",
            "run_directory": "./",
            "config_table": path,
            "share_repo": True,
        }
    )


def test_text_table_vars_valid(config_data_one, text_table_basic):
    member = MemberLevelVars(config_data=config_data_one["member_level"])
    ensemble = EnsembleLevelVars(**_ensemble_cfg(text_table_basic))

    obj = TextTableVars(member_level=member, ensemble_level=ensemble)
    assert isinstance(obj, TextTableVars)


def test_text_table_vars_invalid(config_data_one):
    """Passing an int for ensemble_level should trigger pydantic ValidationError."""
    member = MemberLevelVars(config_data=config_data_one["member_level"])
    with pytest.raises(ValidationError):
        TextTableVars(member_level=member, ensemble_level=int)  # type: ignore[arg-type]


def test_read_table_basic(config_data_one, text_table_basic):
    expected = OmegaConf.create(
        {
            "setup": {
                "runid": ["run-01", "run-02"],
                "exp": ["cmip6-piControl", "cmip6-piControl"],
            }
        }
    )

    member = MemberLevelVars(config_data=config_data_one["member_level"])
    ensemble = EnsembleLevelVars(**_ensemble_cfg(text_table_basic))

    tbl = TextTableVars(member_level=member, ensemble_level=ensemble)

    with pytest.warns(DeprecationWarning):
        result = tbl.read_table()

    assert result == expected


def test_read_table_nested_keys(config_data_one, text_table_nested):
    expected = OmegaConf.create(
        {
            "setup": {
                "runid": ["run-01", "run-02"],
                "exp": ["cmip6-piControl", "cmip6-piControl"],
            },
            "one": {"two": {"three": [1, 2]}},
        }
    )

    member = MemberLevelVars(config_data=config_data_one["member_level"])
    ensemble = EnsembleLevelVars(**_ensemble_cfg(text_table_nested))

    tbl = TextTableVars(member_level=member, ensemble_level=ensemble)
    with pytest.warns(DeprecationWarning):
        result = tbl.read_table()

    assert result == expected


def test_read_table_conflict_raises(config_data_one, text_table_conflict):
    member = MemberLevelVars(config_data=config_data_one["member_level"])
    ensemble = EnsembleLevelVars(**_ensemble_cfg(text_table_conflict))

    tbl = TextTableVars(member_level=member, ensemble_level=ensemble)

    with pytest.raises(ValueError):
        tbl.read_table()



# Pure test to see if we can use the config_manager to do something.
# i.e. create a way to use the config manager in downstream utilities

import copy

from imsi.tools.time_manager.time_manager import sim_time_factory, SimStartStop
from imsi.utils.dict_tools import flatten, replace_curlies_in_dict
from imsi.utils.general import get_date_string


def set_shell_config(shell_config_template: dict, full_config_dict: dict) -> dict:
    """
    Set shell parameters based on shell_config.yaml and internal imsi variables

    Parameters:
        shell_config_template : dict
            A template of the shell config, from the database (contains {{}} enclosed strings to be replaced)

        full_config_dict : dict
            The full configuration as a single dictionary, which will be searched for key=value in order to
            do the replacements in the template above
    """
    # This should possibly be in the config_manager, as technically it is setting a config, not
    # just creating the shell interface. On the other hand, this is pretty specific to the
    # parameters required by shell downstream. More generall the broader
    # use of templates (beyond just shell_config) could be explored. Implicitly other "shell_" functions
    # are defining a type of template (e.g. for computational_env), but `shell_config` is the
    # only template explicitly defined in the imsi upstream configuration database.

    # This provides a general way to create a shell parameters file, built up from variables
    # defined somewhere in the imsi configuration. Its reliance on unique keys could be problematic, and the parsing
    # might lead to issues. Works in basic testing, however the robustness will need to be assessed in the wild.
    # Since we will set shell_config below, we need to first clear it, because it might already contain info
    # raw mapping of imsi to shell variables defined in shell_config.yaml
    # i.e. raw_shell_config is a template, containing {{var}}, where <var> is to be replaced
    raw_shell_config = copy.deepcopy(shell_config_template)
    full_config_dict = copy.deepcopy(full_config_dict) # avoid accidentally modifying upstream mutable dict

    sequencing_config = full_config_dict.get('sequencing')
    local_shell_config = {} # tmp structure to collect replacements
    # For each line of the template, look for {{variable}} to be replaced.
    local_shell_config = replace_curlies_in_dict(raw_shell_config, full_config_dict)

    # This adds selected information from each component to the list of shell parameters, include exe names and MPI sizes.
    # Having this accessing of the values inside the dict reflects substantial coupling!
    for component, component_config in full_config_dict['components'].items():
        if "exec" in component_config.keys():
            local_shell_config[f'{component}_EXEC'] = component_config["exec"]
        if "resources" in component_config.keys():
            # assumes both mpiprocs and ompthreads are present
            local_shell_config[f'{component}_MPIPROCS'] = component_config["resources"]["mpiprocs"]
            local_shell_config[f'{component}_OMPTHREADS'] = component_config["resources"]["ompthreads"]
        if "config" in component_config.keys():
            local_shell_config[f'{component}_CONFIG'] = component_config["config"]
        if "env_params" in component_config:
            local_shell_config.update(component_config["env_params"])

    # Use the time_manager module to get detailed timers for insertion into shell_parameters
    for timer in ['model_submission_job', 'model_inner_loop', 'postproc_submission_job']:
        sim_timer = sim_time_factory(sequencing_config['run_dates'], timer)
        tvars = sim_timer.shell_formatted_time_vars()
        local_shell_config.update(tvars)

    sim_start_stop = SimStartStop.from_kwargs(**sequencing_config['run_dates'])
    tvars = sim_start_stop.shell_formatted_time_vars()
    local_shell_config.update(tvars)

    return local_shell_config


def generate_shell_parameters(shell_config: dict):
    """
    This creates a shell_parameters file contents for the simulation, that contains
    variable definitions required by downstream shell scripting. What appears in
    this file is defined by the "shell_config" template in the imsi json.

    Here we just generate and return a list of strings, that will be written
    to file elsewhere using a common tool.
    """
    shell_config = copy.deepcopy(shell_config)

    shell_parameters_content = list()
    shell_parameters_content.append("# Imsi created shell environment file\n")
    for k, v in shell_config.items():
        shell_parameters_content.append(f'export {k}="{v}"')
    return shell_parameters_content


def generate_flattened_config(full_config_dict: dict):
    """
    This creates a far more extenives shell parameters file content for the simulation, that contains
    every variable defined in the imsi configuration.

    The idea is not to practically use this, but to demonstrate it could be possible
    to make a clean break and do anything in shell downstream of this.

    Here we just generate and return a list of strings, that will be written
    to file elsewhere using a common tool.
    """
    flattened_config = flatten(copy.deepcopy(full_config_dict))
    flattened_config_contents = list()

    flattened_config_contents.append("# Imsi created file\n")
    datestr = get_date_string()
    #f.write(f"# Created for the compiler: {self.compiler} on machine: {self.machine} on date: {datestr}\n")
    for k,v in flattened_config.items():
        flattened_config_contents.append(f'export {k}={v}')

    return flattened_config_contents



from imsi.utils.general import change_dir, delete_or_abort, get_active_venv
from imsi.utils import git_tools
from imsi.user_interface.ui_manager import (
    create_imsi_configuration,
    build_run_config_on_disk,
    validate_version_reqs
)

from pydantic import BaseModel, field_validator, Field, model_validator
from typing import Optional
from pathlib import Path
import shutil
import sys
import os
import subprocess
import sys
import re
import warnings


# runid can contain one or more of the following characters:
#   - Lowercase letters (`a-z`).
#   - Digits (`0-9`).
#   - Hyphens (`-`).
RUNID_PATTERN = re.compile(r"^[a-z0-9-]+$")


class InvalidSetupConfig(Exception):
    """Used to catch invalid setup configurations and provide
    useful feedback to user"""

    pass


class SetupConfigWarning(UserWarning):
    pass


class GitRemoteRepositoryValidator(BaseModel):
    repo_url: str
    remote_branch: Optional[str] = None

    @field_validator("repo_url", mode="after")
    def is_valid_repo_url(url: str) -> bool:
        """Check if a given URL is a valid Git repository."""
        result = subprocess.run([
                    "git", "ls-remote", url
                    ], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        if result.returncode != 0:
            raise InvalidSetupConfig(f"Invalid Git repository URL: '{url}'.")
        return url

    @model_validator(mode="after")
    def is_valid_remote_ref(self):
        """Validate that self.remote_branch is a valid branch or tag in the remote repo.
        Skip validation if it looks like a SHA-1 hash."""

        # Skip validation for anything that looks like a SHA-1
        if git_tools.is_sha1(self.remote_branch):
            return self

        # Check if 'branch' is in the remote branches/tags
        proc = subprocess.run([
                    'git', 'ls-remote', '--exit-code', self.repo_url,
                    '--heads', '--tags', self.remote_branch
                    ],
                    capture_output=True
                    )
        if proc.returncode != 0:
            raise InvalidSetupConfig(
                f"'{self.remote_branch}' is not a valid remote branch or tag in {self.repo_url}."
            )

        return self


class SetupSourceOptions(BaseModel):
    repo: str
    ver: Optional[str] = None
    fetch_method: str

    @field_validator("fetch_method")
    def validate_fetch_method(cls, value):
        valid_fetch = ["clone", "clone-full", "link", "copy"]
        if value not in valid_fetch:
            raise InvalidSetupConfig(
                f"'{value}' is not a valid fetch method. Must in {valid_fetch}."
            )
        return value

    @model_validator(mode="after")
    def error_if_version_unused(self):
        if self.fetch_method in ["clone", "clone-full"] and self.ver is None:
            raise InvalidSetupConfig(
                "\n\n**ERROR**: When --fetch_method is clone, --ver must be specified."
            )
        return self

    @model_validator(mode="after")
    def error_if_version_used(self):
        if self.fetch_method in ["copy", "link"] and self.ver is not None:
            raise InvalidSetupConfig(
                "\n\n**ERROR**: When --fetch_method is copy or link, --ver cannot be specified."
            )
        return self

    @field_validator("repo")
    def validate_git_or_path(cls, value):
        if not Path(value).exists() and not re.match(git_tools.GIT_URL_PATTERN, value):
            raise InvalidSetupConfig(
                f"Check that the repository or path exists: {value}. Ensure remote patterns match: git@<remote>:<project>/<repo> or is a valid URL if using https."
            )
        else:
            return value

    @model_validator(mode="after")
    def error_if_bad_git(self):
        if self.fetch_method in ["clone", "clone-full"]:
            GitRemoteRepositoryValidator(repo_url=self.repo, remote_branch=self.ver)
        if (self.fetch_method == "clone") and git_tools.is_sha1(self.ver):
            # using a hash for setting up a repo as a shallow clone (fetch_method=clone)
            # is problematic for repos with submodules; use the full depth clone
            # instead
            warnings.warn(
                "Can't use --fetch_method=clone and --ver=sha1; continuing with --fetch_method=clone-full",
                SetupConfigWarning
                )
            self.fetch_method = 'clone-full'
        return self


def overwrite_or_abort(path: str, force: bool):
    """Overwrite path if exists or abort.

    The path is relative to cwd. This can be useful for run setup
    folders, where the path is equivalent to the runid.
    """
    if Path(path).exists():
        if force:
            shutil.rmtree(path)
        else:
            print(f"\n\n**WARNING**: The directory for {Path(path).name} already exists.")
            delete_or_abort(path)


def get_imsi_config_path(path):
    imsi_config_path = Path(path) / "imsi-config"
    if not imsi_config_path.exists():
        raise FileNotFoundError(
                f"\n\n **ERROR**: 'imsi-config' directory not found at the top level of the repo provided."
        )
    return imsi_config_path


class ValidatedSetupOptions(SetupSourceOptions):
    """Validate setup inputs early in single location."""

    runid: str = Field(..., max_length=21)
    exp: str | None
    model: Optional[str]
    seq: Optional[str] = None
    machine: Optional[str] = None
    flow: Optional[str] = None
    compiler: Optional[str] = None
    postproc: Optional[str] = None

    @field_validator("runid", mode="before")
    def validate_runid(cls, value):
        if RUNID_PATTERN.match(value) and len(value) < 20:
            return value
        elif len(value) >= 20:
            raise InvalidSetupConfig(
                f"Your runid ---> {value} <--- is too long at {len(value)} chars! It must be less than 20 characters."
            )

        raise InvalidSetupConfig(
            f"Your runid ---> {value} <--- contains unsupported characters! "
            "You must use a runid containing only lowercase alphanumeric characters 'a-z0-9', or hyphens '-'"
            "Also, remember to make it short (<20chars) and unique to you"
            "Examples: runid=ncs-tst-ctrl-01"
        )


def get_source(fetch_method, repo, ver=None,
               path: str = None, source_name: str | None = None,
               force: bool = False, strict: bool = False, verbose: bool = False
               ):
    """Get source repository.

    Inputs
    ------
    fetch_method: str [clone, clone-full, copy, link]
        Fetch method for source code
    repo: str
        URL or path to repository of source code.
    ver : str, None
        Version of source code (for clone* fetch methods only), i.e. the
        git branch.
    path: str, Path
        Base path to where the source folder is placed
    source_name: str, Path
        Name of destination folder in the base path. If None, uses the
        name of the repo (if clone*) or basename of the source folder
        (if copy|link).
    force: bool
        Overrides interactive prompt if the folder (path/source_name)
        already exists.
    strict: bool
        If True, will raise a FileNotFoundError if the 'imsi-config' is
        missing at the top-level of the folder. If False, prints a warning.
    verbose: bool
        Display verbose information. This will display the output from
        git operations.

    Outputs
    -------
        Returns the path to the source that was requested.
    """
    quiet = not verbose

    try:
        src_args = SetupSourceOptions(fetch_method=fetch_method, ver=ver, repo=repo)
    except InvalidSetupConfig as e:
        raise e

    # now use the src_args values, as they have been validated

    if path is None:
        path = Path.cwd()

    if source_name is None:
        if src_args.fetch_method in ['clone', 'clone-full']:
            source_name = git_tools.get_repo_name(repo)
        elif src_args.fetch_method in ['copy', 'link']:
            source_name = Path(src_args.repo).name

    target = path / Path(source_name)

    # make sure that target is not the same as source repo (copy/link)
    if src_args.fetch_method in ['copy', 'link']:
        if target.resolve().absolute() == Path(src_args.repo).resolve().absolute():
            sys.exit(
                f"**ERROR**: source and destination repository represent the same path: {target}"
            )

    overwrite_or_abort(target, force)

    os.makedirs(path, exist_ok=True)

    # Clone, soft link or copy the source code
    if src_args.fetch_method == "clone":
        print(f"Cloning (shallow) {src_args.ver} from {src_args.repo}")
        git_tools.clone(src_args.repo, src_args.ver,
                                       local_name=source_name, path=path,
                                       depth=1, quiet=quiet)
    elif src_args.fetch_method == 'clone-full':
        print(f"Cloning {src_args.ver} from {src_args.repo}")
        git_tools.clone(src_args.repo, src_args.ver,
                                       local_name=source_name, path=path,
                                       depth=None, quiet=quiet)
    elif src_args.fetch_method == "link":
        print(f"Soft linking source from {src_args.repo}")
        os.symlink(Path(src_args.repo).resolve(), target)
    elif src_args.fetch_method == "copy":
        print(f"Copying source from {src_args.repo}")
        shutil.copytree(src_args.repo, target, symlinks=True, ignore_dangling_symlinks=True)

    try:
        get_imsi_config_path(target)
    except FileNotFoundError:
        if strict:
            raise FileNotFoundError
        else:
            print(f"\n**WARNING**: The source does not contain an imsi-config folder under: {target}\n")

    return target


def setup_run(setup_args: ValidatedSetupOptions, force: bool = False, verbose=False, **kwargs):
    """
    Create a run directory, clone in the source code, and checkout version ver

    Inputs:
    -------
    setup_args : ValidatedSetupOptions
        A validated set of setup options for the run
    force: bool
        Force the operation (if the folder exists, delete it, then run the operation).
    verbose : bool
        Display verbose information. This will display the output from
        git operations.

    Outputs:
    --------
        Creates the run directory (named for setup_args.runid) where the
        source (src) contains the contents of the setup_args.repo using
        setup_args.ver and resolved for the imsi setup parameters.
    """

    move_src_to_storage = not kwargs.pop('no_src_storage', False)

    overwrite_or_abort(setup_args.runid, force)

    work_dir = Path(setup_args.runid).resolve()
    work_dir.mkdir(parents=True, exist_ok=True)

    super_repo_source = 'src'   # strict requirement

    src_dir = get_source(
        setup_args.fetch_method, setup_args.repo,
        ver=setup_args.ver, path=work_dir,
        source_name=super_repo_source,
        verbose=verbose,
        force=True,          # always True, due to overwrite_or_abort
        strict=False         # handle this below
        )

    # Requirement that imsi-config appears at the highest setup_args.repo
    try:
        imsi_config_path = get_imsi_config_path(src_dir)
    except FileNotFoundError as e:
        sys.exit(e)

    validate_version_reqs(imsi_config_path)

    imsi_venv = get_active_venv()

    setup_params = {
        "model_name": setup_args.model,               # user
        "experiment_name": setup_args.exp,            # user
        "machine_name": setup_args.machine,           # user - optional
        "compiler_name": setup_args.compiler,         # user - optional
        "runid": setup_args.runid,                    # user
        "work_dir": str(work_dir),                    # user - implicit
        "source_path": str(src_dir),                  # convention
        "source_repo": setup_args.repo,               # tracking
        "source_version": setup_args.ver,             # tracking
        "run_config_path": str(work_dir / "config"),  # convention
        "imsi_config_path": str(imsi_config_path),    # convention
        "fetch_method": setup_args.fetch_method,      # tracking
        "sequencer_name": setup_args.seq,             # user - optional
        "flow_name": setup_args.flow,                 # user - optional
        "imsi_venv": imsi_venv,                       # functional
        "postproc_profile": setup_args.postproc       # user - optional
    }

    with change_dir(work_dir):
        configuration, db = create_imsi_configuration(imsi_config_path, setup_params)

        try:
            src_storage_dir = configuration.machine.setup.src_storage_dir
        except AttributeError:
            src_storage_dir = None

        if move_src_to_storage:
            if (src_storage_dir is not None) and (setup_args.fetch_method != 'link'):
                # move the src to the desired location, link back to local /src

                src_storage_exp = os.path.realpath(os.path.expandvars(src_storage_dir))
                src_storage_runid = os.path.join(src_storage_exp, setup_args.runid)
                ss = os.path.join(src_storage_runid, super_repo_source)

                # but only create the src_storage_dir if it is not the same
                # path as the src under the setup folder
                if Path(ss).resolve() != Path(src_dir).resolve():
                    if os.path.exists(ss):
                        shutil.rmtree(ss)
                    os.makedirs(src_storage_runid, exist_ok=True)

                    try:
                        shutil.move(setup_params['source_path'], ss)
                    except PermissionError:
                        # src left in place
                        warnings.warn(
                            f"Can't move source to {src_storage_exp}: Permission denied.\n"
                            f"source left in place under {setup_args.runid}/{super_repo_source}"
                            )
                    except OSError as e:
                        raise e
                    else:
                        print(f'Soft linking source {setup_args.runid}/{super_repo_source} to {ss}')
                        os.symlink(ss, setup_params['source_path'])

        build_run_config_on_disk(configuration, db)
        print(
            f"\nIMSI setup complete. You can now: \n\n\t\t cd {setup_args.runid} \n"
            f"to continue with configuration/compilation/submission, see:\n\t\t imsi -h"
        )



import pytest
from pydantic import ValidationError
from imsi.config_manager.schema.sequencing import Sequencing


@pytest.fixture
def valid_sequencing():
    return {
        'run_dates': {
            'run_start_time': '2023-01-01',
            'run_stop_time': '2023-12-31',
            'run_segment_start_time': '2023-01-01',
            'run_segment_stop_time': '2023-12-31',
            'model_chunk_size': '1d',
            'model_internal_chunk_size': '1h',
            'postproc_chunk_size': '1h',
        }
    }


@pytest.fixture
def invalid_sequencing():
    return {
        'run_dates': {
            'run_start_time': '2023-01-01',
            'run_stop_time': '2023-12-31',
            'run_segment_start_time': '2023-01-01',
            'run_segment_stop_time': '2023-12-31',
            # Missing model_chunk_size
            'model_internal_chunk_size': '1h',
            'postproc_chunk_size': '1h',
        }
    }


@pytest.fixture
def invalid_sequencing_types():
    return {
        'run_dates': {
            'run_start_time': 20230101,  # Invalid type (should be str)
            'run_stop_time': '2023-12-31',
            'run_segment_start_time': '2023-01-01',
            'run_segment_stop_time': '2023-12-31',
            'model_chunk_size': 1234324,
            'model_internal_chunk_size': '1h',
            'postproc_chunk_size': '1h',
        }
    }


def test_sequencing_valid_config(valid_sequencing):
    """Test valid Sequencing configuration"""
    sequencing = Sequencing(**valid_sequencing)
    assert sequencing.run_dates.run_start_time == '2023-01-01'
    assert sequencing.run_dates.model_chunk_size == '1d'


def test_sequencing_invalid_config(invalid_sequencing):
    """Test invalid Sequencing configuration"""
    with pytest.raises(ValidationError, match='Field required'):
        Sequencing(**invalid_sequencing)


def test_sequencing_invalid_types(invalid_sequencing_types):
    """Test invalid Sequencing configuration with wrong types"""
    with pytest.raises(ValidationError, match='should be a valid string'):
        Sequencing(**invalid_sequencing_types)


def test_sequencing_extra_fields(valid_sequencing):
    """Test Sequencing configuration with extra fields"""
    valid_sequencing['extra_field'] = 'extra_value'
    sequencing = Sequencing(**valid_sequencing)
    assert sequencing.extra_field == 'extra_value'


def test_datacoerce(valid_sequencing):
    """Test DateCoerce conversion"""
    valid_sequencing['run_dates']['run_start_time'] = 20230101  # Coerce int to str
    sequencing = Sequencing(**valid_sequencing)
    assert sequencing.run_dates.run_start_time == '20230101'
    assert isinstance(sequencing.run_dates.run_start_time, str)



import pytest
from omegaconf import OmegaConf, ListConfig

from imsi.tools.ensemble.table_utils.data_model import MemberLevelVars
from imsi.tools.ensemble.table_utils.table_model import BroadcastedConfigList


@pytest.fixture
def _base_setup(request):
    return {
        "ver": request.config.getoption("--target-branch"),
        "repo": request.config.getoption("--target-repo"),
        "fetch_method": "clone",
        "machine": "ppp6",
        "model": request.config.getoption("--target-model"),
    }


@pytest.fixture
def config_data_one(_base_setup):
    return OmegaConf.create(
        {"member_level": {"setup": {**_base_setup, "runid": ["run1"]}}}
    )


@pytest.fixture
def config_data_two(_base_setup):
    return OmegaConf.create(
        {
            "member_level": {
                "setup": {
                    **_base_setup,
                    "runid": ["run-01", "run-02"],
                }
            }
        }
    )


@pytest.fixture
def config_data_auto_runid(_base_setup):
    return OmegaConf.create(
        {
            "member_level": {
                "setup": {
                    **_base_setup,
                    "model": ["canesm51_p1", "canesm51_p1"],
                    "runid": "auto-gen-runid",
                }
            }
        }
    )


def test_expand_config_basic(config_data_one):
    member = MemberLevelVars(config_data=config_data_one["member_level"])
    bcl = BroadcastedConfigList(member_level=member)

    configs = bcl.expand_config(ensemble_size=1)
    assert isinstance(configs, ListConfig)
    assert len(configs) == 1
    # Whatever naming convention the implementation uses, a non‑empty runid
    # string must be present.
    assert configs[0].setup.runid


def test_expand_config_two_members(config_data_two):
    member = MemberLevelVars(config_data=config_data_two["member_level"])
    bcl = BroadcastedConfigList(member_level=member)

    configs = bcl.expand_config(ensemble_size=2)
    assert isinstance(configs, ListConfig)
    assert len(configs) == 2
    # All runids must be unique
    runids = [cfg.setup.runid for cfg in configs]
    assert len(set(runids)) == 2


def test_expand_config_auto_generated_runids(config_data_auto_runid):
    member = MemberLevelVars(config_data=config_data_auto_runid["member_level"])

    # The constructor (and expand_config) emit a warning in this scenario
    bcl = BroadcastedConfigList(member_level=member)
    configs = bcl.expand_config(ensemble_size=2)

    runids = [cfg.setup.runid for cfg in configs]

    assert len(runids) == 2
    assert all(str(rid).startswith("auto-gen-runid") for rid in runids)



import click
import sys

from imsi.tools.simple_sequencer.iss_globals import (
    ALLOWED_JOB_STATUSES,
    JOB_STATUS_FILENAME_TEMPLATE,
    TIME_CHUNK_INDICES
)


def _input_check_timestr_fmt(ctx, param, value):
    import imsi.tools.simple_sequencer.iss as ssi
    pprint = ctx.params.get('print_all')  # for handling mutex args
    if pprint is None or pprint is False:
        if value is not None:
            try:
                ssi.to_unaware_datetime_str(value)
            except Exception as e:
                raise ValueError("input time string must be in format: YYYYmmDDHHMMSS or iso8601") from e
            return value
        else:
            return value
    else:
        return value


@click.group(short_help="imsi simple sequencer (ISS).",
             help="""imsi simple sequencer (ISS).

             ISS is a portable batch sequencing tool. Job scripts are generated
             based on IMSI configuration for jobs, and submitted sequentially
             to the system's scheduler. Its CLI commands are used within the job
             scripts themselves. The user may also use them for further
             development and debugging.
             """,
            context_settings={'show_default': True})
def iss():
    pass

@iss.command(help="Set the status of a job (a specific instance).",
             epilog="A job's status file is stored under the `tracking_dir` with the format: {}".format(JOB_STATUS_FILENAME_TEMPLATE))
@click.option('-j', '--job-name', required=True,
              help="Job name.")
@click.option('-t', '--status', required=True, type=click.Choice(list(ALLOWED_JOB_STATUSES.values()), case_sensitive=True),
              help="Job status (valid iss status option).")
@click.option('-s', '--start-time', required=True, callback=_input_check_timestr_fmt,
              help="Start time string (format: YYYYmmDDHHMMSS or iso8601).")
@click.option('-e', '--stop-time', required=True, callback=_input_check_timestr_fmt,
              help="Stop time string (format: YYYYmmDDHHMMSS or iso8601).")
@click.option('-o', '--tracking-dir', required=False, default=None, type=click.Path(),
              help="Path to tracking directory, ie. where status files are accessed (usually under the run's `/sequencer/listings`)")
def set_status(status, job_name, start_time, stop_time, tracking_dir):
    import imsi.tools.simple_sequencer.iss as ssi

    ssi.set_job_status(job_name, status, start_time, stop_time, tracking_dir)


@iss.command(short_help="Get the current status of a job (a specific instance).",
             help="Get the current status of a job (a specific instance).",
             epilog="A job's status file is stored under the `tracking_dir` with the format: {}".format(JOB_STATUS_FILENAME_TEMPLATE))
@click.option('-j', '--job-name', required=True,
              help="Job name.")
@click.option('-s', '--start-time', required=False, callback=_input_check_timestr_fmt,
              help="Start time string (format: YYYYmmDDHHMMSS or iso8601).")
@click.option('-e', '--stop-time', required=False, callback=_input_check_timestr_fmt,
              help="Start time string (format: YYYYmmDDHHMMSS or iso8601).")
@click.option('-o', '--tracking-dir', required=False, default=None, type=click.Path(),
              help="Path to tracking directory, ie. where status files are accessed (usually under the run's `/sequencer/listings`)")
@click.option('--print-all', is_flag=True,
              help="print all statuses for the job in the tracking dir provided")
def get_status(job_name, start_time, stop_time, tracking_dir, print_all):
    '''cli options and mutex (mutually exclusive) arguemnts handled here, within
    option callback, and option params (note required=False)'''

    import imsi.tools.simple_sequencer.iss as ssi

    mutex_args = {'start_time': start_time, 'stop_time': stop_time}  # w/ print_all
    if print_all:
        for k,v in mutex_args.items():
            if v is not None:
                click.echo(f"INFO: ignored --{k} {v}")
        click.echo(ssi.get_job_status_summary_message(job_name, tracking_dir=tracking_dir))
    else:
        # extra check for cleaner user message
        for k,v in mutex_args.items():
            if v is None:
                raise click.BadParameter(f"--{k} is required")
        click.echo(ssi.get_job_status(job_name, start_time, stop_time, tracking_dir=tracking_dir))

@iss.command(short_help="Get current timer information for the timer.",
             help="""Get current timer information for the timer.

             By default, this function returns all timer elements, where the
             element and it's value are separated by an equal sign (=), and
             multiple elements are separated by a single comma.
             """)
@click.option('-n', '--timer-name', required=True,
              help="Timer name.")
@click.option('-i', '--tracking-file', required=True, default=".simulation.time.state", type=click.Path(exists=True),
              help="Timer tracking file (state file).")
@click.option('--element', required=False, default=None, type=click.Choice(list(TIME_CHUNK_INDICES.keys())))
def get_timer_state(timer_name, tracking_file, element):
    import imsi.tools.time_manager.chunk_manager as cm

    timer_info = cm.get_current_chunk(tracking_file, timer_name)
    if element is None:
        click.echo(",".join(["{}={}".format(k, v) for k,v in timer_info.items()]))
    else:
        click.echo(timer_info[element])

@iss.command(short_help="Submit the job to the scheduler.",
             help="""Submit the job to the scheduler.

             iss internally uses the scheduler's submission command.
             """,
             epilog="The submit command will (naively) submit the job specified. It does not check for up/down stream dependencies or current job or timer state.")
@click.option('-j', '--job-name', required=True,
              help="Job name.")
@click.option('-c', '--cfg-file', required=True, default='.iss.json', type=click.Path(exists=True),
              help="The iss configuration file (json).")
def submit(job_name, cfg_file):
    import imsi.tools.simple_sequencer.iss as ssi

    sequencer = ssi.IMSISimpleSequencer.load_from_json(cfg_file)
    sequencer.run(job_name)

@iss.command(short_help="Determine if the job is ready to submit to the scheduler.",
             help="""Determine if the job is ready to submit to the scheduler.

             A job is deemed "ready to submit" when the job status and timer state
             are resolved and the dependency criteria are met. Further information
             are provided in the documentation.

             Returns:

                 0 : the job is NOT ready to submit.

                 1 : the job is ready to submit (state and dependency conditions are met).
             """
            )
@click.option('-j', '--job-name', required=True,
              help="Job name.")
@click.option('-c', '--cfg-file', required=True, default='.iss.json', type=click.Path(exists=True),
              help="The iss configuration file (json).")
def ready_to_submit(job_name, cfg_file):
    import imsi.tools.simple_sequencer.iss as ssi

    proceed = ssi.ready_to_submit(job_name, cfg_file)
    if proceed:
        click.echo(1)
    else:
        click.echo(0)

@iss.command(short_help="Return a list of dependent job names for the job specified.",
             help="""Return a list of dependent job names for the job specified.

             If there are dependent jobs for the (parent) job specified, the list of
             job names will be separated by a single space.
             """)
@click.option('-j', '--job-name', required=True,
              help="Job name (the parent).")
@click.option('-c', '--cfg-file', required=True, default='.iss.json', type=click.Path(exists=True),
              help="The iss configuration file (json).")
def get_job_dependents(job_name, cfg_file):
    import imsi.tools.simple_sequencer.iss as ssi

    sequencer = ssi.IMSISimpleSequencer.load_from_json(cfg_file)
    if job_name not in sequencer.flow.jobs:
        click.echo(f"job '{job_name}' not found in current configuration: {cfg_file}", err=True)
        sys.exit(1)
    if job_name in sequencer.flow.job_dependents:
        click.echo(' '.join([j.job_name for j in sequencer.flow.job_dependents[job_name]]))
    else:
        # the job doesnt have dependents - nothing to return
        sys.exit(0)

@iss.command(help="Get the timer name for the job specified.")
@click.option('-j', '--job-name', required=True,
              help="Job name.")
@click.option('-c', '--cfg-file', required=True, default='.iss.json', type=click.Path(exists=True),
              help="The iss configuration file (json).")
def get_timer_name(job_name, cfg_file):
    import imsi.tools.simple_sequencer.iss as ssi

    timer_name = ssi.get_timer_name(job_name, cfg_file)
    click.echo(timer_name)

@iss.command(help="Query information for argument specified.")
@click.argument('arg', type=click.Choice(['jobs', 'scheduler', 'submit']))
@click.option('-c', '--cfg-file', required=False, default='.iss.json', type=click.Path(exists=True, dir_okay=False),
              help="The iss configuration file (json).")
def which(arg, cfg_file):
    import imsi.tools.simple_sequencer.iss as ssi

    if arg == "jobs":
        click.echo(ssi.which_jobs(cfg_file))
    elif arg == "scheduler":
        click.echo(ssi.which_scheduler(cfg_file))
    elif arg == "submit":
        click.echo(ssi.which_submit(cfg_file))

if __name__ == "__main__":
    iss()



import pytest
from click.testing import CliRunner
from imsi.cli.entry import cli
from pathlib import Path
import subprocess
import shutil
import os

from imsi.user_interface.setup_manager import InvalidSetupConfig

@pytest.fixture
def config_params(request):
    pars = {
        "repo": request.config.getoption("--target-repo"),
        "branch": request.config.getoption("--target-branch"),
        "model": request.config.getoption("--target-model"),
        "experiment": request.config.getoption("--target-exp"),
        "machine": request.config.getoption("--target-machine"),
        "seq": request.config.getoption("--target-sequencer"),
    }
    return pars


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def local_test_dir(request):
    """Create a temporary test directory under the current working directory and clean up after."""
    tmp_dir = Path.cwd() / f"test_tmp_{request.node.name}"
    if tmp_dir.exists():
        shutil.rmtree(tmp_dir)
    tmp_dir.mkdir(parents=True)
    yield tmp_dir
    shutil.rmtree(tmp_dir, ignore_errors=True)


def clone_repo(target_repo, target_branch, tmp_dir, local_repo='local_repo'):
    """Generic clone command for testing."""
    subprocess.run(
        [
            "git", "clone",
            "--depth", "1",
            "--shallow-submodules",
            "--branch", target_branch,
            target_repo,
            local_repo,
        ],
        check=True,
        cwd=Path(tmp_dir).resolve(),
    )

    repo_path = Path(tmp_dir, local_repo).resolve()
    subprocess.run(["git", "submodule", "init", "CanNEMO"], check=True, cwd=repo_path)
    subprocess.run(["git", "submodule", "update", "--init", "--recursive", "CanNEMO"], check=True, cwd=repo_path)


def run_test(
    runner,
    fetch_method,
    tmp_dir,
    target_repo,
    target_model,
    target_experiment,
    target_machine,
    target_sequencer,
    target_branch=None,
):

    runid = "test-run-tmp"

    copy_link_cmd = [
        "-f", "setup",
        f"--runid={runid}",
        f"--repo={target_repo}",
        f"--fetch_method={fetch_method}",
        f"--model={target_model}",
        f"--exp={target_experiment}",
        f"--machine={target_machine}",
        f"--seq={target_sequencer}",
        "--no-src-storage"  # this needs to be here for unit tests due to conflicting runner/default storage dirs
    ]
    clone_cmd = [
        "-f", "setup",
        f"--runid={runid}",
        f"--repo={target_repo}",
        "--fetch_method=clone",
        f"--model={target_model}",
        f"--ver={target_branch}",
        f"--exp={target_experiment}",
        f"--machine={target_machine}",
        f"--seq={target_sequencer}",
        "--no-src-storage" # this needs to be here for unit tests due to conflicting runner/default storage dirs
    ]

    # 👇 manually change into tmp_dir
    cwd_before = os.getcwd()
    os.chdir(tmp_dir)
    try:
        result = runner.invoke(
            cli,
            copy_link_cmd if fetch_method in ["copy", "link"] else clone_cmd,
            catch_exceptions=False,
        )
    finally:
        os.chdir(cwd_before)

    try:
        test_run_dir = Path(tmp_dir) / runid
        assert result.exit_code == 0, result.output
        assert test_run_dir.exists(), f"Run directory not found: {test_run_dir}"
        assert (test_run_dir / ".imsi-setup.log").exists(), "Log file not found"
        assert (test_run_dir / "src").exists(), "Source code not found"
    except Exception as e:
        print(f"Error occurred: {e}")
        shutil.rmtree(tmp_dir, ignore_errors=True)

def test_setup_tree_clone(config_params, runner, local_test_dir):
    run_test(
        runner=runner,
        fetch_method="clone",
        tmp_dir=local_test_dir,
        target_repo=config_params["repo"],
        target_model=config_params["model"],
        target_experiment=config_params["experiment"],
        target_machine=config_params["machine"],
        target_sequencer=config_params["seq"],
        target_branch=config_params["branch"],
    )


@pytest.mark.parametrize("fetch_method", ["copy", "link"])
def test_setup_tree_copy_link(config_params, runner, local_test_dir, fetch_method):
    local_repo_name = "local_repo"
    repo_path = local_test_dir / local_repo_name

    clone_repo(
        target_repo=config_params["repo"],
        target_branch=config_params["branch"],
        tmp_dir=local_test_dir,
        local_repo=local_repo_name,
    )

    run_test(
        runner=runner,
        fetch_method=fetch_method,
        tmp_dir=local_test_dir,
        target_repo=repo_path.resolve(),
        target_model=config_params["model"],
        target_experiment=config_params["experiment"],
        target_machine=config_params["machine"],
        target_sequencer=config_params["seq"],
    )


def test_setup_invalid_runid(config_params, runner):

    # required args
    target_repo=config_params['repo']
    target_model=config_params['model']
    target_exp=config_params['experiment']
    bad_runid = "test_run"   # underscore not allowed

    # will fail
    with pytest.raises(InvalidSetupConfig):
        runner.invoke(
            cli,
            [
                "-f", "setup", f"--runid={bad_runid}",
                f"--repo={target_repo}", f"--model={target_model}",
                f"--exp={target_exp}"
            ],
            catch_exceptions=False
            )



import pytest
from pydantic import ValidationError
from imsi.config_manager.schema.compiler import Compiler


def test_minimal_config():
    """Test minimal configuration of Compiler."""
    config = Compiler(name='test_compiler')
    assert config.name == 'test_compiler'

    with pytest.raises(ValidationError):
        Compiler()
